"""Module to validate indicator source data before uploading to the public COVIDcast API.

This file defines the functions that are made public by the module. As the
module is intended to be executed though the main method, these are primarily
for testing.
"""



"""Read parameter files containing configuration information."""
# -*- coding: utf-8 -*-
from json import load,dump
from shutil import copyfile, move
import os
import sys

def read_params():
    """Read a file named 'params.json' in the current working directory.

    If the file does not exist, it copies the file 'params.json.template' to
    'params.json' and then reads the file.
    """
    if not os.path.exists("params.json"):
        copyfile("params.json.template", "params.json")

    with open("params.json", "r") as json_file:
        return load(json_file)

def params_get(path, params):
    """Read a key/path from loaded params.

    Parameters
    ----------
    path : str
      The .-delimited path to the value to return. Currently supports dict keys
      but not numeric array indices.

    params : dict
      Params as read using read_params().
    """
    r = params
    for p in path.split("."):
        r = r[p]
    return r

def params_set(path, value, params):
    """Set a key/path in loaded params.

    Parameters
    ----------
    path : str
      The .-delimited path to the value to set. Currently supports dict keys
      but not numeric array indices.

    value : str
      The value to assign at the specified location. Supports several types:
        "true" "false" - assign boolean
        ,-delimited value, or
        existing value is list - assign list
        "/dev/fd/..." - assign string by reading path as file
        anything else - assign string

    params : dict
      Params as read using read_params().
    """
    path = path.split(".")
    for p in path[:-1]:
        params = params[p] # stick the 'params' handle on the innermost dict/list
    if value.startswith("/dev/fd/"): # explicitly handle process substitutions
        with open(value) as f:
            value = f.read().strip()
    if value in {"true","false"}:
        value = value == "true"
    elif isinstance(params[path[-1]],list) or value.find(",")>0:
        value = value.split(",")
    params[path[-1]] = value

def params_run():
    """Get or set parameter value from command line arguments."""
    if len(sys.argv)<3:
        print("""
Usage:
        python -m delphi_utils set key1 value1 key2 value2a,value2b,value2c [...]
        python -m delphi_utils get key1.key2.key3[...]
""")
        sys.exit()

    params = read_params()
    if sys.argv[1]=="get":
        print(params_get(sys.argv[2], params))
    if sys.argv[1]=="set":
        n=0
        for k,v in zip(sys.argv[2::2], sys.argv[3::2]):
            params_set(k, v, params)
            n += 1
        with open("params.json", "w") as f:
            dump(params, f, sort_keys=True, indent=2)
        print(f"Updated {n} items")

def transfer_files():
    """Transfer files to prepare for acquisition."""
    params = read_params()
    export_dir = params["common"].get("export_dir", None)
    delivery_dir = params["delivery"].get("delivery_dir", None)
    files_to_export = os.listdir(export_dir)
    for file_name in files_to_export:
        if file_name.lower().endswith(".csv"):
            move(os.path.join(export_dir, file_name), os.path.join(delivery_dir, file_name))

def delete_move_files():
    """Delete or move output files depending on dir settings provided in params.

    1. Delete files in export-dir if delivery-dir is specified and is different
       from export_dir (aka only delete files produced by the most recent run)
    2. If validation-failures-dir is specified, move failed files there instead
    If dry-run tag is True, then this function should not (and currently does not) get called
    """
    params = read_params()
    export_dir = params["common"].get("export_dir", None)
    delivery_dir = params["delivery"].get("delivery_dir", None)
    validation_failure_dir = params["validation"]["common"].get("validation_failure_dir", None)
    # Create validation_failure_dir if it doesn't exist
    if (validation_failure_dir is not None) and (not os.path.exists(validation_failure_dir)):
        os.mkdir(validation_failure_dir)
    # Double-checking that export-dir is not delivery-dir
    # Throw assertion error if delivery_dir or export_dir is unspecified
    assert(delivery_dir is not None and export_dir is not None)
    assert export_dir != delivery_dir
    files_to_delete = os.listdir(export_dir)
    for file_name in files_to_delete:
        if file_name.endswith(".csv") or file_name.endswith(".CSV"):
            if validation_failure_dir is not None:
                move(os.path.join(export_dir, file_name),
                    os.path.join(validation_failure_dir, file_name))
            else:
                os.remove(os.path.join(export_dir, file_name))



"""Slack notifier for use in Delphi indicators."""
from slack import WebClient

class SlackNotifier:
    """Notifies slack channels of messages."""

    def __init__(self, slack_channel, slack_token):
        """Initialize with a slack channel and token."""
        self.slack_channel = slack_channel
        self.client = WebClient(token = slack_token)

    def post_message(self, blocks):
        """Post a message to the slack channel."""
        self.client.chat_postMessage(
            channel=self.slack_channel,
            blocks=blocks
        )



"""Structured logger utility for creating JSON logs in Delphi pipelines."""
import logging
import sys
import threading
import structlog


def handle_exceptions(logger):
    """Handle exceptions using the provided logger."""
    def exception_handler(etype, value, traceback):
        logger.exception("Top-level exception occurred",
                         exc_info=(etype, value, traceback))

    def multithread_exception_handler(args):
        exception_handler(args.exc_type, args.exc_value, args.exc_traceback)

    sys.excepthook = exception_handler
    threading.excepthook = multithread_exception_handler


def get_structured_logger(name=__name__,
                          filename=None,
                          log_exceptions=True):
    """Create a new structlog logger.

    Use the logger returned from this in indicator code using the standard
    wrapper calls, e.g.:

    logger = get_structured_logger(__name__)
    logger.warning("Error", type="Signal too low").

    The output will be rendered as JSON which can easily be consumed by logs
    processors.

    See the structlog documentation for details.

    Parameters
    ---------
    name: Name to use for logger (included in log lines), __name__ from caller
    is a good choice.
    filename: An (optional) file to write log output.
    """
    # Configure the basic underlying logging configuration
    logging.basicConfig(
        format="%(message)s",
        level=logging.INFO,
        handlers=[logging.StreamHandler()]
    )

    # Configure structlog. This uses many of the standard suggestions from
    # the structlog documentation.
    structlog.configure(
        processors=[
            # Filter out log levels we are not tracking.
            structlog.stdlib.filter_by_level,
            # Include logger name in output.
            structlog.stdlib.add_logger_name,
            # Include log level in output.
            structlog.stdlib.add_log_level,
            # Allow formatting into arguments e.g., logger.info("Hello, %s",
            # name)
            structlog.stdlib.PositionalArgumentsFormatter(),
            # Add timestamps.
            structlog.processors.TimeStamper(fmt="iso"),
            # Match support for exception logging in the standard logger.
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            # Decode unicode characters
            structlog.processors.UnicodeDecoder(),
            # Render as JSON
            structlog.processors.JSONRenderer()
        ],
        # Use a dict class for keeping track of data.
        context_class=dict,
        # Use a standard logger for the actual log call.
        logger_factory=structlog.stdlib.LoggerFactory(),
        # Use a standard wrapper class for utilities like log.warning()
        wrapper_class=structlog.stdlib.BoundLogger,
        # Cache the logger
        cache_logger_on_first_use=True,
    )

    # Create the underlying python logger and wrap it with structlog
    system_logger = logging.getLogger(name)
    if filename and not system_logger.handlers:
        system_logger.addHandler(logging.FileHandler(filename))
    system_logger.setLevel(logging.INFO)
    logger = structlog.wrap_logger(system_logger)

    if log_exceptions:
        handle_exceptions(logger)

    return logger



"""Dynamic file checks."""
from dataclasses import dataclass
from datetime import date, timedelta
from typing import Dict, Set
import re
import pandas as pd
import numpy as np
from .errors import ValidationFailure
from .datafetcher import get_geo_signal_combos, threaded_api_calls
from .utils import relative_difference_by_min, TimeWindow, lag_converter


class DynamicValidator:
    """Class for validation of static properties of individual datasets."""

    @dataclass
    class Parameters:
        """Configuration parameters."""

        # data source name, one of
        # https://cmu-delphi.github.io/delphi-epidata/api/covidcast_signals.html
        data_source: str
        # span of time over which to perform checks
        time_window: TimeWindow
        # date that this df_to_test was generated; typically 1 day after the last date in df_to_test
        generation_date: date
        # number of days back to perform sanity checks, starting from the last date appearing in
        # df_to_test
        max_check_lookbehind: timedelta
        # names of signals that are smoothed (7-day avg, etc)
        smoothed_signals: Set[str]
        # maximum number of days behind do we expect each signal to be
        max_expected_lag: Dict[str, int]
        # minimum number of days behind do we expect each signal to be
        min_expected_lag: Dict[str, int]

    def __init__(self, params):
        """
        Initialize object and set parameters.

        Arguments:
            - params: dictionary of user settings; if empty, defaults will be used
        """
        common_params = params["common"]
        dynamic_params = params.get("dynamic", dict())

        self.test_mode = dynamic_params.get("test_mode", False)

        self.params = self.Parameters(
            data_source=common_params["data_source"],
            time_window=TimeWindow.from_params(common_params["end_date"],
                                               common_params["span_length"]),
            generation_date=date.today(),
            max_check_lookbehind=timedelta(
                days=max(7, dynamic_params.get("ref_window_size", 14))),
            smoothed_signals=set(dynamic_params.get("smoothed_signals", [])),
            min_expected_lag=lag_converter(common_params.get(
                "min_expected_lag", dict())),
            max_expected_lag=lag_converter(common_params.get(
                "max_expected_lag", dict()))
        )

    def validate(self, all_frames, report):
        """
        Perform all checks over the combined data set from all files.

        Parameters
        ----------
        all_frames: pd.DataFrame
            combined data from all input files
        report: ValidationReport
            report to which the results of these checks will be added
        """
        # Get 14 days prior to the earliest list date
        outlier_lookbehind = timedelta(days=14)

        # Get all expected combinations of geo_type and signal.
        geo_signal_combos = get_geo_signal_combos(self.params.data_source)

        all_api_df = threaded_api_calls(self.params.data_source,
                                        self.params.time_window.start_date - outlier_lookbehind,
                                        self.params.time_window.end_date,
                                        geo_signal_combos)

        # Keeps script from checking all files in a test run.
        kroc = 0

        # Comparison checks
        # Run checks for recent dates in each geo-sig combo vs semirecent (previous
        # week) API data.
        for geo_type, signal_type in geo_signal_combos:
            geo_sig_df = all_frames.query(
                "geo_type == @geo_type & signal == @signal_type")
            # Drop unused columns.
            geo_sig_df.drop(columns=["geo_type", "signal"])

            report.increment_total_checks()

            if geo_sig_df.empty:
                report.add_raised_error(ValidationFailure(check_name="check_missing_geo_sig_combo",
                                                          geo_type=geo_type,
                                                          signal=signal_type,
                                                          message="file with geo_type-signal combo "
                                                                  "does not exist"))
                continue

            max_date = geo_sig_df["time_value"].max()
            self.check_min_allowed_max_date(
                max_date, geo_type, signal_type, report)
            self.check_max_allowed_max_date(
                max_date, geo_type, signal_type, report)

            self.check_na_vals(geo_sig_df, geo_type, signal_type, report)

            # Get relevant reference data from API dictionary.
            api_df_or_error = all_api_df[(geo_type, signal_type)]

            report.increment_total_checks()
            if not isinstance(api_df_or_error, pd.DataFrame):
                report.add_raised_error(api_df_or_error)
                continue

            # Only do outlier check for cases and deaths signals
            if (signal_type in ["confirmed_7dav_cumulative_num", "confirmed_7dav_incidence_num",
                                "confirmed_cumulative_num", "confirmed_incidence_num",
                                "deaths_7dav_cumulative_num",
                                "deaths_cumulative_num"]):
                # Outlier dataframe
                earliest_available_date = geo_sig_df["time_value"].min()
                source_df = geo_sig_df.query(
                    'time_value <= @self.params.time_window.end_date & '
                    'time_value >= @self.params.time_window.start_date'
                )

                # These variables are interpolated into the call to `api_df_or_error.query()`
                # below but pylint doesn't recognize that.
                # pylint: disable=unused-variable
                outlier_start_date = earliest_available_date - outlier_lookbehind
                outlier_end_date = earliest_available_date - timedelta(days=1)
                outlier_api_df = api_df_or_error.query(
                    'time_value <= @outlier_end_date & time_value >= @outlier_start_date')
                # pylint: enable=unused-variable

                self.check_positive_negative_spikes(
                    source_df, outlier_api_df, geo_type, signal_type, report)

            # Check data from a group of dates against recent (previous 7 days,
            # by default) data from the API.
            for checking_date in self.params.time_window.date_seq:
                create_dfs_or_error = self.create_dfs(
                    geo_sig_df, api_df_or_error, checking_date, geo_type, signal_type, report)

                if not create_dfs_or_error:
                    continue
                recent_df, reference_api_df = create_dfs_or_error

                self.check_max_date_vs_reference(
                    recent_df, reference_api_df, checking_date, geo_type, signal_type, report)

                self.check_rapid_change_num_rows(
                    recent_df, reference_api_df, checking_date, geo_type, signal_type, report)

                if not re.search("cumulative", signal_type):
                    self.check_avg_val_vs_reference(
                        recent_df, reference_api_df, checking_date, geo_type,
                        signal_type, report)

            # Keeps script from checking all files in a test run.
            kroc += 1
            if self.test_mode and kroc == 2:
                break

    def check_na_vals(self, geo_sig_df, geo_type, signal_type, report):
        """Check if there are any NA values.

        In particular, make sure that error doesn't occur for new Geo IDs introduced.

        Arguments:
            - geo_type: str; geo type name (county, msa, hrr, state) as in the CSV name
            - signal_type: str; signal name as in the CSV name
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        def replace_first_six(df, start_date):
            x = df.val.isnull()
            # First 6 days have to be null
            x.iloc[:6] = False
            df = df[x]
            return df.time_value[df.time_value >= start_date]

        grouped_df = geo_sig_df.groupby('geo_id')
        error_df = grouped_df.apply(replace_first_six,
            start_date = self.params.time_window.start_date)

        if not error_df.empty:
            for index, value in error_df.iteritems():
                report.add_raised_error(
                    ValidationFailure("check_val_missing",
                                      geo_type=geo_type,
                                      signal=signal_type,
                                      date=value,
                                      message=f"geo_id {index[0]}"
                                      )
                )

        report.increment_total_checks()

    def check_min_allowed_max_date(self, max_date, geo_type, signal_type, report):
        """Check if time since data was generated is reasonable or too long ago.

        The most recent data should be at least max_expected_lag from generation date

        Arguments:
            - max_date: date of most recent data to be validated; datetime format.
            - geo_type: str; geo type name (county, msa, hrr, state) as in the CSV name
            - signal_type: str; signal name as in the CSV name
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        min_thres = timedelta(days = self.params.max_expected_lag.get(
            signal_type, self.params.max_expected_lag.get('all', 10)))

        if max_date < self.params.generation_date - min_thres:
            report.add_raised_error(
                ValidationFailure("check_min_max_date",
                                  geo_type=geo_type,
                                  signal=signal_type,
                                  message="date of most recent generated file seems too long ago"))

        report.increment_total_checks()

    def check_max_allowed_max_date(self, max_date, geo_type, signal_type, report):
        """Check if time since data was generated is reasonable or too recent.

        The most recent data should be at most min_expected_lag from generation date

        Arguments:
            - max_date: date of most recent data to be validated; datetime format.
            - geo_type: str; geo type name (county, msa, hrr, state) as in the CSV name
            - signal_type: str; signal name as in the CSV name
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        max_thres = timedelta(days = self.params.min_expected_lag.get(
            signal_type, self.params.min_expected_lag.get('all', 1)))

        if max_date > self.params.generation_date - max_thres:
            report.add_raised_error(
                ValidationFailure("check_max_max_date",
                                  geo_type=geo_type,
                                  signal=signal_type,
                                  message="date of most recent generated file seems too recent"))

        report.increment_total_checks()

    def create_dfs(self, geo_sig_df, api_df_or_error, checking_date, geo_type, signal_type, report):
        """Create recent_df and reference_api_df from params.

        Raises error if recent_df is empty.

        Arguments:
            - geo_sig_df: Pandas dataframe of test data
            - api_df_or_error: pandas dataframe of reference data, either from the
            COVIDcast API or semirecent data
            - geo_type: str; geo type name (county, msa, hrr, state) as in the CSV name
            - signal_type: str; signal name as in the CSV name
            - report: ValidationReport; report where results are added

        Returns:
            - False if recent_df is empty after padding, else (recent_df, reference_api_df)
            (reference_api_df will be padded if necessary)
        """
        # recent_lookbehind: start from the check date and working backward in time,
        # how many days at a time do we want to check for anomalies?
        # Choosing 1 day checks just the daily data.
        recent_lookbehind = timedelta(days=1)

        recent_cutoff_date = checking_date - \
            recent_lookbehind + timedelta(days=1)
        recent_df = geo_sig_df.query(
            'time_value <= @checking_date & time_value >= @recent_cutoff_date')

        report.increment_total_checks()

        if recent_df.empty:
            min_thres = timedelta(days = self.params.max_expected_lag.get(
                signal_type, self.params.max_expected_lag.get('all', 10)))
            if checking_date < self.params.generation_date - min_thres:
                report.add_raised_error(
                    ValidationFailure("check_missing_geo_sig_date_combo",
                                    checking_date,
                                    geo_type,
                                    signal_type,
                                    "test data for a given checking date-geo type-signal type"
                                    " combination is missing. Source data may be missing"
                                    " for one or more dates"))
            return False

        # Reference dataframe runs backwards from the recent_cutoff_date
        #
        # These variables are interpolated into the call to `api_df_or_error.query()`
        # below but pylint doesn't recognize that.
        # pylint: disable=unused-variable
        reference_start_date = recent_cutoff_date - self.params.max_check_lookbehind
        if signal_type in self.params.smoothed_signals:
            # Add an extra 7 days to the reference period.
            reference_start_date = reference_start_date - \
                timedelta(days=7)

        reference_end_date = recent_cutoff_date - timedelta(days=1)
        # pylint: enable=unused-variable

        # Subset API data to relevant range of dates.
        reference_api_df = api_df_or_error.query(
            "time_value >= @reference_start_date & time_value <= @reference_end_date")

        pre_pad_empty_flag = reference_api_df.empty
        reference_api_df = self.pad_reference_api_df(
            reference_api_df, geo_sig_df, reference_start_date, reference_end_date)

        report.increment_total_checks()
        if reference_api_df.empty:
            report.add_raised_error(ValidationFailure("empty_reference_data",
                                  checking_date,
                                  geo_type,
                                  signal_type,
                                  "reference data is empty; comparative checks could not "
                                  "be performed"))
            return False
        if pre_pad_empty_flag:
            report.add_raised_warning(ValidationFailure("empty_reference_data",
                                  checking_date,
                                  geo_type,
                                  signal_type,
                                  "pre-padding reference data is empty and indicates data "
                                  "missing from the API; please verify that this is expected"))

        return (geo_sig_df, reference_api_df)

    # `reference_start_date` is used in the call to `geo_sig_df.query()`
    # below but pylint doesn't recognize that.
    # pylint: disable=unused-argument
    def pad_reference_api_df(self, reference_api_df, geo_sig_df,
                             reference_start_date, reference_end_date):
        """Check if API data is missing, and supplement from test data.

        Arguments:
            - reference_api_df: API data within lookbehind range
            - geo_sig_df: Test data
            - reference_start_date: Desired start date of reference data
            - reference_end_date: Desired end date of reference data

        Returns:
            - reference_api_df: Supplemented version of original
        """
        # Value is `NaT` (not a time) if reference_api_df is empty.
        reference_api_df_max_date = reference_api_df.time_value.max().date()
        if reference_api_df.empty:
            geo_sig_df_supplement = geo_sig_df.query(
                'time_value <= @reference_end_date & time_value >= \
                @reference_start_date')[[
                "geo_id", "val", "se", "sample_size", "time_value"]]
        elif reference_api_df_max_date < reference_end_date:
            # If actual end date `reference_api_df_max_date` is not as recent as
            # the desired end date `reference_end_date`, add rows from recently
            # generate data, in `geo_sig_df`, to the reference data.
            geo_sig_df_supplement = geo_sig_df.query(
                'time_value <= @reference_end_date & time_value > \
                @reference_api_df_max_date')[[
                "geo_id", "val", "se", "sample_size", "time_value"]]
        else:
            return reference_api_df

        # Final processing after supplementing reference_api_df
        # Matching time_value format
        geo_sig_df_supplement["time_value"] = \
            pd.to_datetime(geo_sig_df_supplement["time_value"],
                format = "%Y-%m-%d %H:%M:%S")
        reference_api_df = pd.concat(
            [reference_api_df, geo_sig_df_supplement])

        return reference_api_df

    def check_max_date_vs_reference(self, df_to_test, df_to_reference, checking_date,
                                    geo_type, signal_type, report):
        """
        Check if reference data is more recent than test data.

        Arguments:
            - df_to_test: pandas dataframe of a single CSV of source data
            (one day-signal-geo_type combo)
            - df_to_reference: pandas dataframe of reference data, either from the
            COVIDcast API or semirecent data
            - geo_type: str; geo type name (county, msa, hrr, state) as in the CSV name
            - signal_type: str; signal name as in the CSV name
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        if df_to_test["time_value"].max() < df_to_reference["time_value"].max():
            report.add_raised_error(
                ValidationFailure("check_max_date_vs_reference",
                                  checking_date,
                                  geo_type,
                                  signal_type,
                                  "reference df has days beyond the max date in the =df_to_test="))

        report.increment_total_checks()

    def check_rapid_change_num_rows(self, df_to_test, df_to_reference, checking_date,
                                    geo_type, signal_type, report):
        """
        Compare number of obervations per day in test dataframe vs reference dataframe.

        Arguments:
            - df_to_test: pandas dataframe of CSV source data
            - df_to_reference: pandas dataframe of reference data, either from the
            COVIDcast API or semirecent data
            - checking_date: datetime date
            - geo_type: str; geo type name (county, msa, hrr, state) as in the CSV name
            - signal_type: str; signal name as in the CSV name
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        test_rows_per_reporting_day = df_to_test[df_to_test['time_value']
                                                 == checking_date].shape[0]
        reference_rows_per_reporting_day = df_to_reference.shape[0] / len(
            set(df_to_reference["time_value"]))

        try:
            compare_rows = relative_difference_by_min(
                test_rows_per_reporting_day,
                reference_rows_per_reporting_day)
        except ZeroDivisionError as e:
            print(checking_date, geo_type, signal_type)
            raise e

        if abs(compare_rows) > 0.35:
            report.add_raised_error(
                ValidationFailure("check_rapid_change_num_rows",
                                  checking_date,
                                  geo_type,
                                  signal_type,
                                  "Number of rows per day seems to have changed rapidly (reference "
                                  "vs test data)"))
        report.increment_total_checks()

    def check_positive_negative_spikes(self, source_df, api_frames, geo, sig, report):
        """
        Adapt Dan's corrections package to Python (only consider spikes).

        See https://github.com/cmu-delphi/covidcast-forecast/tree/dev/corrections/data_corrections

        Statistics for a right shifted rolling window and a centered rolling window are used
        to determine outliers for both positive and negative spikes.

        As it is now, ststat will always be NaN for source frames.

        Arguments:
            - source_df: pandas dataframe of CSV source data
            - api_frames: pandas dataframe of reference data, either from the
            COVIDcast API or semirecent data
            - geo: str; geo type name (county, msa, hrr, state) as in the CSV name
            - sig: str; signal name as in the CSV name
            - report: ValidationReport; report where results are added

        """
        report.increment_total_checks()
        # Combine all possible frames so that the rolling window calculations make sense.
        source_frame_start = source_df["time_value"].min()
        # This variable is interpolated into the call to `add_raised_error()`
        # below but pylint doesn't recognize that.
        # pylint: disable=unused-variable
        source_frame_end = source_df["time_value"].max()
        # pylint: enable=unused-variable
        all_frames = pd.concat([api_frames, source_df]). \
            drop_duplicates(subset=["geo_id", "time_value"], keep='last'). \
            sort_values(by=['time_value']).reset_index(drop=True)

        # Tuned Variables from Dan's Code for flagging outliers. Size_cut is a
        # check on the minimum value reported, sig_cut is a check
        # on the ftstat or ststat reported (t-statistics) and sig_consec
        # is a lower check for determining outliers that are next to each other.
        size_cut, sig_cut, sig_consec = 5, 3, 2.25

        # Functions mapped to rows to determine outliers based on fstat and ststat values

        def outlier_flag(frame):
            if (abs(frame["val"]) > size_cut) and not (pd.isna(frame["ststat"])) \
                    and (frame["ststat"] > sig_cut):
                return True
            if (abs(frame["val"]) > size_cut) and (pd.isna(frame["ststat"])) and \
                    not (pd.isna(frame["ftstat"])) and (frame["ftstat"] > sig_cut):
                return True
            if (frame["val"] < -size_cut) and not (pd.isna(frame["ststat"])) and \
                    not pd.isna(frame["ftstat"]):
                return True
            return False

        def outlier_nearby(frame):
            if (not pd.isna(frame['ststat'])) and (frame['ststat'] > sig_consec):
                return True
            if pd.isna(frame['ststat']) and (frame['ftstat'] > sig_consec):
                return True
            return False

        # Calculate ftstat and ststat values for the rolling windows, group fames by geo region
        region_group = all_frames.groupby("geo_id")
        window_size = 14
        # Shift the window to match how R calculates rolling windows with even numbers
        shift_val = -1 if window_size % 2 == 0 else 0

        # Calculate the t-statistics for the two rolling windows (windows center and windows right)
        all_full_frames = []
        for _, group in region_group:
            rolling_windows = group["val"].rolling(
                window_size, min_periods=window_size)
            center_windows = group["val"].rolling(
                window_size, min_periods=window_size, center=True)
            fmedian = rolling_windows.median()
            smedian = center_windows.median().shift(shift_val)
            fsd = rolling_windows.std() + 0.00001  # if std is 0
            ssd = center_windows.std().shift(shift_val) + 0.00001  # if std is 0
            group['ftstat'] = abs(group["val"] - fmedian.fillna(0)) / fsd
            group['ststat'] = abs(group["val"] - smedian.fillna(0)) / ssd
            all_full_frames.append(group)

        all_frames = pd.concat(all_full_frames)
        # Determine outliers in source frames only, only need the reference
        # data from just before the start of the source data
        # because lead and lag outlier calculations are only one day
        #
        # These variables are interpolated into the call to `api_df_or_error.query()`
        # below but pylint doesn't recognize that.
        # pylint: disable=unused-variable
        api_frames_end = min(api_frames["time_value"].max(),
                             source_frame_start-timedelta(days=1))
        # pylint: enable=unused-variable
        outlier_df = all_frames.query(
            'time_value >= @api_frames_end & time_value <= @source_frame_end')
        outlier_df = outlier_df.sort_values(by=['geo_id', 'time_value']) \
            .reset_index(drop=True).copy()
        outliers = outlier_df[outlier_df.apply(outlier_flag, axis=1)]
        outliers_reset = outliers.copy().reset_index(drop=True)

        # Find the lead outliers and the lag outliers. Check that the selected row
        # is actually a leading and lagging row for given geo_id
        upper_index = list(filter(lambda x: x < outlier_df.shape[0],
                                  list(outliers.index+1)))
        upper_df = outlier_df.iloc[upper_index, :].reset_index(drop=True)
        upper_compare = outliers_reset[:len(upper_index)]
        sel_upper_df = upper_df[upper_compare["geo_id"]
                                == upper_df["geo_id"]].copy()
        lower_index = list(filter(lambda x: x >= 0, list(outliers.index-1)))
        lower_df = outlier_df.iloc[lower_index, :].reset_index(drop=True)
        # If lower_df is empty, then make lower_compare empty too
        if lower_df.empty:
            lower_compare = outliers_reset[0:0]
        else:
            lower_compare = outliers_reset[-len(lower_index):].reset_index(drop=True)
        sel_lower_df = lower_df[lower_compare["geo_id"]
                                == lower_df["geo_id"]].copy()

        outliers_list = [outliers]
        if sel_upper_df.size > 0:
            outliers_list.append(
                sel_upper_df[sel_upper_df.apply(outlier_nearby, axis=1)])
        if sel_lower_df.size > 0:
            outliers_list.append(
                sel_lower_df[sel_lower_df.apply(outlier_nearby, axis=1)])

        all_outliers = pd.concat(outliers_list). \
            sort_values(by=['time_value', 'geo_id']). \
            drop_duplicates().reset_index(drop=True)

        # Identify outliers just in the source data
        source_outliers = all_outliers.query(
            "time_value >= @source_frame_start & time_value <= @source_frame_end")

        if source_outliers.shape[0] > 0:
            for time_val in source_outliers["time_value"].unique():
                report.add_raised_warning(
                    ValidationFailure(
                        "check_positive_negative_spikes",
                        time_val,
                        geo,
                        sig,
                        "Source dates with flagged ouliers based on the previous 14 days of data "
                        "available"))

    def check_avg_val_vs_reference(self, df_to_test, df_to_reference, checking_date, geo_type,
                                   signal_type, report):
        """
        Compare average values for each variable in test dataframe vs reference dataframe.

        Arguments:
            - df_to_test: pandas dataframe of CSV source data
            - df_to_reference: pandas dataframe of reference data, either from the
            COVIDcast API or semirecent data
            - geo_type: str; geo type name (county, msa, hrr, state) as in the CSV name
            - signal_type: str; signal name as in the CSV name
            - report: ValidationReport; report where results are added
        Returns:
            - None
        """
        # Calculate reference mean and standard deviation for each geo_id.
        reference_mean = df_to_reference.groupby(['geo_id'], as_index=False)[
            ['val', 'se', 'sample_size']].mean().assign(type="reference mean")
        reference_sd = df_to_reference.groupby(['geo_id'], as_index=False)[
            ['val', 'se', 'sample_size']].std().round(8).assign(type="reference sd")
        reference_count = df_to_reference.groupby(['geo_id'], as_index=False)[
            ['val', 'se', 'sample_size']].count().assign(type="reference count")

        # Replace standard deviations of 0 with non-zero min sd for that type. Ignores NA.
        replacements = {"val": {0: reference_sd.val[reference_sd.val > 0].median()},
                        "se": {0: reference_sd.se[reference_sd.se > 0].median()},
                        "sample_size": {0: reference_sd.sample_size[
                            reference_sd.sample_size > 0].median()}}
        reference_sd.replace(replacements, inplace=True)

        # Duplicate reference_mean and reference_sd for every unique time_value seen in df_to_test
        reference_df = pd.concat(
            [reference_mean, reference_sd, reference_count]
        ).assign(
            key=0
        ).merge(
            df_to_test.assign(key=0)[["time_value", "key"]].drop_duplicates()
        ).drop("key", axis=1)

        # Drop unused columns from test data.
        df_to_test = df_to_test[[
            "geo_id", "val", "se", "sample_size", "time_value"
        ]].assign(
            type="test"
        )

        # For each variable (val, se, and sample size) where not missing, calculate the
        # mean z-score and mean absolute z-score of the test data across all geographic
        # regions and dates.
        #
        # Approach:
        #  - Use each reference df to calculate mean and sd for each geo_id (above). Merge
        #    onto test data.
        #  - Use to calculate z-score for each test datapoint for a given geo_id and date.
        #  - Avg z-scores over each geo_id, across all dates.
        #  - Avg all z-scores together.
        num_ref_dates = self.params.max_check_lookbehind.days
        if signal_type in self.params.smoothed_signals:
            num_ref_dates += 7

        df_all = pd.concat(
            [df_to_test, reference_df]
        ).melt(
            id_vars=["geo_id", "type", "time_value"], value_vars=["val", "se", "sample_size"]
        ).pivot(
            index=("geo_id", "variable", "time_value"), columns="type", values="value"
        ).reset_index(
            ("geo_id", "variable", "time_value")
        ).dropna(
        ).assign(
            z=lambda x: (
                x["test"] - x["reference mean"]) / x["reference sd"],
            abs_z=lambda x: abs(x["z"])
        ).replace([np.inf, -np.inf], np.nan, inplace = False
        ).query("`reference count` == @num_ref_dates"
        ).dropna(
        ).groupby(
            ["geo_id", "variable"], as_index=False
        ).agg(
            geo_z=("z", "mean"),
            geo_abs_z=("abs_z", "mean")
        ).groupby(
            "variable", as_index=False
        ).agg(
            mean_z=("geo_z", "mean"),
            mean_abs_z=("geo_abs_z", "mean")
        )[["variable", "mean_z", "mean_abs_z"]]

        # Set thresholds for comparison.
        classes = ['mean_z', 'val_mean_z', 'mean_abs_z']
        thres = pd.DataFrame([[4.0, 3.5, 4.25]], columns=classes)

        # Check if the calculated mean differences are high compared to the thresholds.
        mean_z_high = (
            abs(df_all["mean_z"]) > float(thres["mean_z"])).any() or (
                (df_all["variable"] == "val").any() and
                (abs(df_all[df_all["variable"] == "val"]["mean_z"])
                 > float(thres["val_mean_z"])).any()
        )
        mean_abs_z_high = (df_all["mean_abs_z"] > float(
            thres["mean_abs_z"])).any()

        if mean_z_high or mean_abs_z_high:
            report.add_raised_warning(
                ValidationFailure(
                    "check_test_vs_reference_avg_changed",
                    checking_date,
                    geo_type,
                    signal_type,
                    'Average differences in variables by geo_id between recent & reference data '
                    + 'seem large --- either large increase '
                    + 'tending toward one direction or large mean absolute difference, relative '
                    + 'to average values of corresponding variables. For the former check, '
                    + 'tolerances for `val` are more restrictive than those for other columns.'))

        report.increment_total_checks()



"""
Utilities for diffing and archiving covidcast export CSVs.

Aims to simplify the creation of issues for new and backfilled value for indicators.
Also handles archiving of export CSVs to some backend (git, S3 etc.) before replacing them.

Example workflow regardless of specific ArchiveDiffer used. Should only differ in intialization.
1) Initialize and update cache folder if neccessary
>>> arch_diff = S3ArchiveDiffer(cache_dir, export_dir, ...)
>>> arch_diff.update_cache()
>>> ... # Run indicator and generate full exports in `export_dir`

2) Create new diff files from cache files vs export files
>>> deleted_files, common_diffs, new_files = arch_diff.diff_exports()

3) Archive common files with diffs and new files
>>> to_archive = [f for f, diff in common_diffs.items() if diff is not None]
>>> to_archive += new_files
>>> succs, fails = arch_diff.archive_exports(to_archive)

4) Filter exports: Replace files with their diffs, or remove if no diffs
>>> succ_common_diffs = {f: diff for f, diff in common_diffs.items() if f not in fails}
>>> arch_diff.filter_exports(succ_common_diffs)

Author: Eu Jing Chua
Created: 2020-08-06
"""

from contextlib import contextmanager
import filecmp
from glob import glob
from os import remove, replace
from os.path import join, basename, abspath
import shutil
import time
from typing import Tuple, List, Dict, Optional

from boto3 import Session
from boto3.exceptions import S3UploadFailedError
from git import Repo
from git.refs.head import Head
import pandas as pd
import numpy as np

from .utils import read_params
from .logger import get_structured_logger
from .nancodes import Nans

Files = List[str]
FileDiffMap = Dict[str, Optional[str]]

EXPORT_CSV_DTYPES = {
    "geo_id": str, "val": float, "se": float, "sample_size": float,
    "missing_val": "Int64", "missing_se": "Int64", "missing_sample_size": "Int64"
}

def diff_export_csv(
    before_csv: str,
    after_csv: str
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Find differences in exported covidcast CSVs, using geo_id as the index.

    The signal and standard error values are rounded to 7 decimal places before comparison.

    Treats NA == NA as True.

    Parameters
    ----------
    before_csv: str
        The CSV file to diff from
    after_csv: str
        The CSV file to diff to

    Returns
    -------
        (deleted_df, changed_df, added_df)
        deleted_df is the pd.DataFrame of deleted rows from before_csv.
        changed_df is the pd.DataFrame of common rows from after_csv with changed values.
        added_df is the pd.DataFrame of added rows from after_csv.
    """
    before_df = pd.read_csv(before_csv, dtype=EXPORT_CSV_DTYPES)
    before_df.set_index("geo_id", inplace=True)
    before_df = before_df.round({"val": 7, "se": 7})
    after_df = pd.read_csv(after_csv, dtype=EXPORT_CSV_DTYPES)
    after_df.set_index("geo_id", inplace=True)
    after_df = after_df.round({"val": 7, "se": 7})
    deleted_idx = before_df.index.difference(after_df.index)
    common_idx = before_df.index.intersection(after_df.index)
    added_idx = after_df.index.difference(before_df.index)

    before_df_cmn = before_df.reindex(common_idx)
    after_df_cmn = after_df.reindex(common_idx)

    # If new CSV has missingness columns, but old doesn't, mark all values as new
    if ("missing_val" not in before_df_cmn.columns) & ("missing_val" in after_df_cmn.columns):
        same_mask = after_df_cmn.copy()
        same_mask.loc[:] = False
    else:
        # Exact comparisons, treating NA == NA as True
        same_mask = before_df_cmn == after_df_cmn
        same_mask |= pd.isna(before_df_cmn) & pd.isna(after_df_cmn)

    # Any deleted entries become rows with nans and the deleted missing code
    deleted_df = before_df.loc[deleted_idx, :].copy()
    deleted_df[["val", "se", "sample_size"]] = np.nan
    # If the new file doesn't have missing columsn, then when the deleted, changed, and added
    # rows are concatenated (in diff_exports), they will default to NA
    deleted_df[["missing_val", "missing_se", "missing_sample_size"]] = Nans.DELETED

    return (
        deleted_df,
        after_df_cmn.loc[~(same_mask.all(axis=1)), :],
        after_df.loc[added_idx, :])


def archiver_from_params(params):
    """Build an ArchiveDiffer from `params`.

    The type of ArchiveDiffer constructed is inferred from the parameters.

    Parameters
    ----------
    params: Dict[str, Dict[str, Any]]
        Dictionary of user-defined parameters with the following structure:
        - "common":
            - "export_dir": str, directory to which indicator output files have been exported
        - "archive":
            - "cache_dir": str, directory containing cached data from previous indicator runs
            - "branch_name" (required for git archiver): str, name of git branch
            - "override_dirty" (optional for git archiver): bool, whether to allow overwriting of
                untracked & uncommitted changes in `cache_dir`
            - "commit_partial_success" (optional for git archiver): bool, whether to still commit
                even if some files were not archived and staged due to `override_dirty=False`
            - "commit_message" (optional for git archiver): str, commit message to use
            - "bucket_name" (required for S3 archiver): str, name of S3 bucket to which to upload
                files
            - "indicator_prefix" (required for S3 archiver): str, S3 prefix for files from this
                indicator
            - "aws_credentials" (required for S3 archiver): Dict[str, str], authentication
                parameters for S3 to create a boto3.Session

    Returns
    -------
    ArchiveDiffer of the inferred type.
    """
    if "archive" not in params:
        return None

    # Copy to kwargs to take advantage of default arguments to archiver
    kwargs = params["archive"]
    kwargs["export_dir"] = params["common"]["export_dir"]

    if "branch_name" in kwargs:
        return GitArchiveDiffer(**kwargs)

    if "bucket_name" in kwargs:
        assert "indicator_prefix" in kwargs, "Missing indicator_prefix in params"
        assert "aws_credentials" in kwargs, "Missing aws_credentials in params"
        return S3ArchiveDiffer(**kwargs)

    # Don't run the filesystem archiver if the user misspecified the archiving params
    assert set(kwargs.keys()) == set(["cache_dir", "export_dir"]),\
        'If you intended to run a filesystem archiver, please remove all options other than '\
        '"cache_dir" from the "archive" params.  Otherwise, please include either "branch_name" '\
        'or "bucket_name" to run the git or S3 archivers, respectively.'
    return FilesystemArchiveDiffer(**kwargs)


class ArchiveDiffer:
    """Base class for performing diffing and archiving of exported covidcast CSVs."""

    def __init__(self, cache_dir: str, export_dir: str):
        """
        Initialize an ArchiveDiffer.

        Parameters
        ----------
        cache_dir: str
            The directory for storing most recent archived/uploaded CSVs to do start diffing from.
            Usually 'cache'.
        export_dir: str
            The directory with most recent exported CSVs to diff to.
            Usually 'receiving'.
        """
        self.cache_dir = cache_dir
        self.export_dir = export_dir

        self._cache_updated = False
        self._exports_archived = False

    def update_cache(self):
        """
        Make sure cache_dir is updated correctly from a backend.

        To be implemented by specific archiving backends.
        Should set self._cache_updated = True after verifying cache is updated.
        """
        raise NotImplementedError

    def diff_exports(self) -> Tuple[Files, FileDiffMap, Files]:
        """
        Find diffs across and within CSV files, from cache_dir to export_dir.

        Should be called after update_cache() succeeds. Only works on *.csv files,
        ignores every other file.

        Returns
        -------
        (deleted_files, common_diffs, new_files): Tuple[Files, FileDiffMap, Files]
            deleted_files: List of files that are present in cache_dir but missing in export_dir.
            common_diffs: Dict mapping common files in export_dir with cache_dir to:
                          - None, if the common file is identical
                          - None, if the export_dir version only has DELETED rows
                          - a filename with .csv.diff suffix, containing ADDED and CHANGED rows ONLY
            added_files: List of files that are missing in cache_dir but present in export_dir.
        """
        assert self._cache_updated

        # Glob to only pick out CSV files, ignore hidden files
        previous_files = set(basename(f)
                             for f in glob(join(self.cache_dir, "*.csv")))
        exported_files = set(basename(f)
                             for f in glob(join(self.export_dir, "*.csv")))

        deleted_files = sorted(join(self.cache_dir, f)
                               for f in previous_files - exported_files)
        common_filenames = sorted(exported_files & previous_files)
        new_files = sorted(join(self.export_dir, f)
                           for f in exported_files - previous_files)

        common_diffs: Dict[str, Optional[str]] = {}
        for filename in common_filenames:
            before_file = join(self.cache_dir, filename)
            after_file = join(self.export_dir, filename)

            common_diffs[after_file] = None

            # Check for simple file similarity before doing CSV diffs
            if filecmp.cmp(before_file, after_file, shallow=False):
                continue

            deleted_df, changed_df, added_df = diff_export_csv(
                before_file, after_file)
            new_issues_df = pd.concat([deleted_df, changed_df, added_df], axis=0)

            if len(deleted_df) > 0:
                print(
                    f"Diff has deleted indices in {after_file} that have been coded as nans.")

            # Write the diffs to diff_file, if applicable
            if len(new_issues_df) > 0:
                diff_file = join(self.export_dir, filename + ".diff")

                new_issues_df.to_csv(diff_file, na_rep="NA")
                common_diffs[after_file] = diff_file

        return deleted_files, common_diffs, new_files

    def archive_exports(self, exported_files: Files) -> Tuple[Files, Files]:
        """
        Handle actual archiving of files, depending on specific backend.

        To be implemented by specific archiving backends.

        Parameters
        ----------
        exported_files: Files
            List of files to be archived. Usually new and changed files.

        Returns
        -------
        (successes, fails): Tuple[Files, Files]
            successes: List of successfully archived files
            fails: List of unsuccessfully archived files
        """
        raise NotImplementedError

    def filter_exports(self, common_diffs: FileDiffMap):
        """
        Filter export directory to only contain relevant files.

        Filters down the export_dir to only contain:
        1) New files, 2) Changed files, filtered-down to the ADDED and CHANGED rows only.
        Should be called after archive_exports() so we archive the raw exports before
        potentially modifying them.

        Parameters
        ----------
        common_diffs: FileDiffmap
            Same semantics as in diff_exports(). For each exported_file, diff key-value pair:
                1) If the diff is None, remove exported_file from export_dir
                2) If there is a diff, replace exported_file with the diff
            Since this is done for all key-value pairs, one can filter down common_diffs to only
            a subset of exported_files to operate on before calling filter_exports.
            For example, removing keys that correspond to failed-to-archive files.
        """
        # Should be called after archive_exports
        assert self._exports_archived

        for exported_file, diff_file in common_diffs.items():
            # Delete existing exports that had no data diff
            if diff_file is None:
                remove(exported_file)

            # Replace exports where diff file was generated
            else:
                replace(diff_file, exported_file)

    def run(self, logger=None):
        """Run the differ and archive the changed and new files."""
        start_time = time.time()
        self.update_cache()

        # Diff exports, and make incremental versions
        _, common_diffs, new_files = self.diff_exports()

        # Archive changed and new files only
        to_archive = [f for f, diff in common_diffs.items()
                      if diff is not None]
        if logger:
            logger.debug("Diffed exports",
                         phase="archiving",
                         new_files_count=len(new_files),
                         common_diffs_count=len(to_archive))
        to_archive += new_files
        _, fails = self.archive_exports(to_archive)

        # Filter existing exports to exclude those that failed to archive
        succ_common_diffs = {f: diff for f,
                             diff in common_diffs.items() if f not in fails}
        self.filter_exports(succ_common_diffs)

        # Report failures: someone should probably look at them
        for exported_file in fails:
            print(f"Failed to archive '{exported_file}'")

        elapsed_time_in_seconds = round(time.time() - start_time, 2)
        if logger:
            logger.info("Completed archive run",
                        phase="archiving",
                        elapsed_time_in_seconds=elapsed_time_in_seconds,
                        new_changed_count=len(to_archive),
                        fail_count=len(fails))


class S3ArchiveDiffer(ArchiveDiffer):
    """
    AWS S3 backend for archiving.

    Archives CSV files into a S3 bucket, with keys "{indicator_prefix}/{csv_file_name}".
    Ideally, versioning should be enabled in this bucket to track versions of each CSV file.
    """

    def __init__(
        self, cache_dir: str, export_dir: str,
        bucket_name: str,
        indicator_prefix: str,
        aws_credentials: Dict[str, str],
    ):
        """
        Initialize a S3ArchiveDiffer.

        See this link for possible aws_credentials kwargs:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session

        Parameters
        ----------
        cache_dir: str
            The directory for storing most recent archived/uploaded CSVs to do start diffing from.
            Usually 'cache'.
        export_dir: str
            The directory with most recent exported CSVs to diff to.
            Usually 'receiving'.
        bucket_name: str
            The S3 bucket to upload files to.
        indicator_prefix: str
            The prefix for S3 keys related to this indicator.
        aws_credentials: Dict[str, str]
            kwargs to create a boto3.Session, containing AWS credentials/profile to use.
        """
        super().__init__(cache_dir, export_dir)
        self.s3 = Session(**aws_credentials).resource("s3")
        self.bucket = self.s3.Bucket(bucket_name)
        self.indicator_prefix = indicator_prefix

    def update_cache(self):
        """Make sure cache_dir is updated with all latest files from the S3 bucket."""
        # List all indicator-related objects from S3
        archive_objects = self.bucket.objects.filter(
            Prefix=self.indicator_prefix).all()
        archive_objects = [
            obj for obj in archive_objects if obj.key.endswith(".csv")]

        # Check against what we have locally and download missing ones
        cached_files = set(basename(f)
                           for f in glob(join(self.cache_dir, "*.csv")))
        for obj in archive_objects:
            archive_file = basename(obj.key)
            cached_file = join(self.cache_dir, archive_file)

            if archive_file not in cached_files:
                print(f"Updating cache with {cached_file}")
                obj.Object().download_file(cached_file)

        self._cache_updated = True

    def archive_exports(self,  # pylint: disable=arguments-differ
        exported_files: Files,
        update_cache: bool = True,
        update_s3: bool = True
    ) -> Tuple[Files, Files]:
        """
        Handle actual archiving of files to the S3 bucket.

        Parameters
        ----------
        exported_files: Files
            List of files to be archived. Usually new and changed files.

        Returns
        -------
        (successes, fails): Tuple[Files, Files]
            successes: List of successfully archived files
            fails: List of unsuccessfully archived files
        """
        archive_success = []
        archive_fail = []

        for exported_file in exported_files:
            cached_file = abspath(
                join(self.cache_dir, basename(exported_file)))
            archive_key = join(self.indicator_prefix, basename(exported_file))

            try:
                if update_cache:
                    # Update local cache
                    shutil.copyfile(exported_file, cached_file)

                if update_s3:
                    self.bucket.Object(archive_key).upload_file(exported_file)

                archive_success.append(exported_file)
            except FileNotFoundError:
                archive_fail.append(exported_file)
            except shutil.SameFileError:
                # no need to copy if the cached file is the same
                archive_success.append(exported_file)

        self._exports_archived = True

        return archive_success, archive_fail


class GitArchiveDiffer(ArchiveDiffer):
    """
    Local git repo backend for archiving.

    Archives CSV files into a local git repo as commits.
    Assumes that a git repository is already set up.
    """

    def __init__(
        self, cache_dir: str, export_dir: str,
        branch_name: Optional[str] = None,
        override_dirty: bool = False,
        commit_partial_success: bool = False,
        commit_message: str = "Automated archive",
    ):
        """
        Initialize a GitArchiveDiffer.

        Parameters
        ----------
        cache_dir: str
            The directory for storing most recent archived/uploaded CSVs to do start diffing from.
            Either cache_dir or some parent dir should be a git repository. Usually 'cache'.
        export_dir: str
            The directory with most recent exported CSVs to diff to.
            Usually 'receiving'.
        branch_name: Optional[str]
            Branch to use for archiving. Uses current branch if None.
        override_dirty: bool
            Whether to allow overwriting of untracked & uncommitted changes in cache_dir.
        commit_partial_success: bool
            Whether to still commit even if some files were not archived and staged due
            to override_dirty=False
        commit_message: str
            The automatic commit message to use for the commit.
        """
        super().__init__(cache_dir, export_dir)

        assert override_dirty or not commit_partial_success, \
            "Only can commit_partial_success=True when override_dirty=True"

        # Assumes a repository is set up already, will raise exception if not found
        self.repo = Repo(cache_dir, search_parent_directories=True)

        self.branch = self.get_branch(branch_name)
        self.override_dirty = override_dirty
        self.commit_partial_success = commit_partial_success
        self.commit_message = commit_message

    def get_branch(self, branch_name: Optional[str] = None) -> Head:
        """
        Retrieve a Head object representing a branch of specified name.

        Creates the branch from the current active branch if does not exist yet.

        Parameters
        ----------
        branch_name: Optional[str]
            If None, just returns current branch. Otherwise, retrieves/creates branch.

        Returns
        -------
        branch: Head
        """
        if branch_name is None:
            return self.repo.active_branch

        if branch_name in self.repo.branches:
            return self.repo.branches[branch_name]

        return self.repo.create_head(branch_name)

    @contextmanager
    def archiving_branch(self):
        """
        Context manager for checking out a branch.

        Useful for checking out self.branch within a context, then switching back
        to original branch when finished.
        """
        orig_branch = self.repo.active_branch
        self.branch.checkout()

        try:
            yield self.branch
        finally:
            orig_branch.checkout()

    def update_cache(self):
        """
        Check if cache_dir is clean: has everything nicely committed if override_dirty=False.

        Since we are using a local git repo, assumes there is nothing to update from.
        """
        # Make sure cache directory is clean: has everything nicely committed
        if not self.override_dirty:
            cache_clean = not self.repo.is_dirty(
                untracked_files=True, path=abspath(self.cache_dir))
            assert cache_clean, f"There are uncommitted changes in the cache dir '{self.cache_dir}'"

        self._cache_updated = True

    def diff_exports(self) -> Tuple[Files, FileDiffMap, Files]:
        """
        Find diffs across and within CSV files, from cache_dir to export_dir.

        Same as base class diff_exports, but in context of specified branch.
        """
        with self.archiving_branch():
            return super().diff_exports()

    def archive_exports(self, exported_files: Files) -> Tuple[Files, Files]:
        """
        Handle actual archiving of files to the local git repo.

        Parameters
        ----------
        exported_files: Files
            List of files to be archived. Usually new and changed files.

        Returns
        -------
        (successes, fails): Tuple[Files, Files]
            successes: List of successfully archived files
            fails: List of unsuccessfully archived files
        """
        archived_files = []
        archive_success = []
        archive_fail = []

        with self.archiving_branch():
            # Abs paths of all modified files to check if we will override uncommitted changes
            working_tree_dir = self.repo.working_tree_dir
            dirty_files = [join(working_tree_dir, f)
                           for f in self.repo.untracked_files]
            dirty_files += [join(working_tree_dir, d.a_path)
                            for d in self.repo.index.diff(None)]

            for exported_file in exported_files:
                archive_file = abspath(
                    join(self.cache_dir, basename(exported_file)))

                # Archive and explicitly stage new export, depending if override
                if self.override_dirty or archive_file not in dirty_files:
                    try:
                        # Archive
                        shutil.copyfile(exported_file, archive_file)

                        archived_files.append(archive_file)
                        archive_success.append(exported_file)

                    except (FileNotFoundError, S3UploadFailedError) as ex:
                        print(ex)
                        archive_fail.append(exported_file)

                # Otherwise ignore the archiving for this file
                else:
                    archive_fail.append(exported_file)

            # Stage
            self.repo.index.add(archived_files)

            # Commit staged files
            if len(exported_files) > 0:

                # Support partial success and at least one archive succeeded
                partial_success = self.commit_partial_success and len(
                    archive_success) > 0

                if len(archive_success) == len(exported_files) or partial_success:
                    self.repo.index.commit(message=self.commit_message)

        self._exports_archived = True

        return archive_success, archive_fail

class FilesystemArchiveDiffer(ArchiveDiffer):
    """Filesystem-based backend for archiving.

    This backend is only intended for use to reconstruct historical issues whose
    versioning history has already been tracked elsewhere. No versioning is
    performed by this backend and the cache directory is modified in-place
    without backups. Do not use it for new issues.
    """

    def archive_exports(self, exported_files):
        """Handle file archiving with a no-op.

        FilesystemArchiveDiffer does not track versioning information, so this
        just does enough to convince the caller that it's okay to proceed with
        the next step.

        Parameters
        ----------
        exported_files: Files
            List of files to be archived. Usually new and changed files.

        Returns
        -------
        (successes, fails): Tuple[Files, Files]
            successes: All files from input
            fails: Empty list
        """
        archive_success, archive_fail = [], []
        for exported_file in exported_files:
            archive_file = abspath(
                join(self.cache_dir, basename(exported_file)))

            # Copy export to cache
            try:
                # Archive
                shutil.copyfile(exported_file, archive_file)
                archive_success.append(exported_file)

            except FileNotFoundError as ex:
                print(ex)
                archive_fail.append(exported_file)

        self._exports_archived = True
        return archive_success, archive_fail

    def update_cache(self):
        """Handle cache updates with a no-op.

        FilesystemArchiveDiffer does not track versioning information, so this
        just does enough to convince the caller that it's okay to proceed with
        the next step.
        """
        self._cache_updated = True

if __name__ == "__main__":
    _params = read_params()

    # Autodetect whether parameters have been factored hierarchically or not
    # See https://github.com/cmu-delphi/covidcast-indicators/issues/847
    # Once all indicators have their parameters factored in to "common", "indicator", "validation",
    # and "archive", this code will be obsolete.
    #
    # We assume that by virtue of invoking this module from the command line that the user intends
    # to run validation.  Thus if the "archive" sub-object is not found, we interpret that to mean
    # the parameters have not be hierarchically refactored.
    if "archive" not in _params:
        _params = {"archive": _params, "common": _params}

    archiver_from_params(_params).run(get_structured_logger(
        __name__, filename=_params["common"].get("log_filename"),
        log_exceptions=_params["common"].get("log_exceptions", True)))



# -*- coding: utf-8 -*-
"""Call the function params_run when executed.

This file indicates that calling the module (`python -m delphi_utils`) will
call the function `params_run` found within the utils.py file. There should be
no need to change this template.
"""

from .utils import params_run # pragma: no cover

params_run() # pragma: no cover



"""Unified not-a-number codes for CMU Delphi codebase."""

from enum import IntEnum

class Nans(IntEnum):
    """An enum of not-a-number codes for the indicators."""

    NOT_MISSING = 0
    NOT_APPLICABLE = 1
    REGION_EXCEPTION = 2
    CENSORED = 3
    DELETED = 4
    OTHER = 5



"""Validation output reports."""
import sys
from typing import List
from ..logger import get_structured_logger
from .errors import ValidationFailure

class ValidationReport:
    """Class for reporting the results of validation."""

    # pylint: disable=R0902
    def __init__(self, errors_to_suppress: List[ValidationFailure],
                 data_source: str = "", dry_run: bool = False):
        """Initialize a ValidationReport.

        Parameters
        ----------
        errors_to_suppress: List[ValidationFailure]
            List of ValidationFailures to ignore.
        data_source: str
            Name of data source as obtained from params

        Attributes
        ----------
        errors_to_suppress: List[ValidationFailure]
            See above
        data_source: str
            See above
        num_suppressed: int
            Number of errors suppressed
        total_checks: int
            Number of validation checks performed
        raised_errors: List[Exception]
            Errors raised from validation failures
        raised_warnings: List[Exception]
            Warnings raised from validation execution
        unsuppressed_errors: List[Exception]
            Errors raised from validation failures not found in `self.errors_to_suppress`
        """
        self.errors_to_suppress = errors_to_suppress
        self.data_source = data_source
        self.num_suppressed = 0
        self.total_checks = 0
        self.raised_errors = []
        self.raised_warnings = []
        self.unsuppressed_errors = []
        self.dry_run = dry_run
    # pylint: enable=R0902

    def add_raised_error(self, error):
        """Add an error to the report.

        Parameters
        ----------
        error: Exception
            Error raised in validation

        Returns
        -------
        None
        """
        self.raised_errors.append(error)
        if error.is_suppressed(self.errors_to_suppress):
            self.num_suppressed += 1
        else:
            self.unsuppressed_errors.append(error)

    def increment_total_checks(self):
        """Record a check."""
        self.total_checks += 1

    def add_raised_warning(self, warning):
        """Add a warning to the report.

        Parameters
        ----------
        warning: Warning
            Warning raised in validation

        Returns
        -------
        None
        """
        self.raised_warnings.append(warning)

    def log(self, logger=None):
        """Log errors and warnings."""
        if logger is None:
            logger = get_structured_logger(__name__)

        if self.success():
            logger.info("Validation run successful",
                data_source = self.data_source,
                checks_run = self.total_checks,
                checks_failed = len(self.unsuppressed_errors),
                checks_suppressed = self.num_suppressed,
                warnings = len(self.raised_warnings),
                phase = "validation")
        else:
            logger.info("Validation run unsuccessful",
                data_source = self.data_source,
                checks_run = self.total_checks,
                checks_failed = len(self.unsuppressed_errors),
                checks_suppressed = self.num_suppressed,
                warnings = len(self.raised_warnings),
                phase="validation")
        # Threshold for slack alerts if warnings are excessive,
        # Currently extremely strict, set by observation of 1 month's logs
        excessive_warnings = self.total_checks > 0 and \
            (len(self.raised_warnings) > 200 or \
            len(self.raised_warnings) / self.total_checks > 0.015)
        if excessive_warnings:
            logger.info("Excessive number of warnings",
                data_source = self.data_source,
                checks_run = self.total_checks,
                checks_failed = len(self.unsuppressed_errors),
                checks_suppressed = self.num_suppressed,
                warnings = len(self.raised_warnings),
                phase = "validation")
        for error in self.unsuppressed_errors:
            date_str = "*" if error.date is None else error.date.isoformat()
            logger.critical(str(error),
                phase = "validation",
                error_name = error.check_name,
                signal = error.signal,
                resolution = error.geo_type,
                date = date_str)
        for warning in self.raised_warnings:
            date_str = "*" if warning.date is None else warning.date.isoformat()
            logger.warning(str(warning),
                phase="validation",
                error_name = warning.check_name,
                signal = warning.signal,
                resolution = warning.geo_type,
                date = date_str)

    def print_and_exit(self, logger=None, die_on_failures=True):
        """Print results and exit.

        Arguments
        ---------
        die_on_failures: bool
            Whether to return non-zero status if any failures were encountered.
        """
        self.log(logger)
        if self.success():
            sys.exit(0)
        elif die_on_failures:
            sys.exit(1)

    def success(self):
        """Determine if the report corresponds to a successful validation run."""
        return len(self.unsuppressed_errors) == 0 or self.dry_run



"""Export data in the format expected by the Delphi API."""
# -*- coding: utf-8 -*-
from datetime import datetime
from os.path import join
from typing import Optional
import logging

from epiweeks import Week
import numpy as np
import pandas as pd

from .nancodes import Nans

def filter_contradicting_missing_codes(df, sensor, metric, date, logger=None):
    """Find values with contradictory missingness codes, filter them, and log."""
    columns = ["val", "se", "sample_size"]
    # Get indicies where the XNOR is true (i.e. both are true or both are false).
    masks = [
        ~(df[column].isna() ^ df["missing_" + column].eq(Nans.NOT_MISSING))
        for column in columns
    ]
    for mask in masks:
        if not logger is None and df.loc[mask].size > 0:
            logger.info(
                "Filtering contradictory missing code in " +
                "{0}_{1}_{2}.".format(sensor, metric, date.strftime(format="%Y-%m-%d"))
            )
            df = df.loc[~mask]
        elif logger is None and df.loc[mask].size > 0:
            df = df.loc[~mask]
    return df

def create_export_csv(
    df: pd.DataFrame,
    export_dir: str,
    geo_res: str,
    sensor: str,
    metric: Optional[str] = None,
    start_date: Optional[datetime] = None,
    end_date: Optional[datetime] = None,
    remove_null_samples: Optional[bool] = False,
    write_empty_days: Optional[bool] = False,
    logger: Optional[logging.Logger] = None,
    weekly_dates = False,
):
    """Export data in the format expected by the Delphi API.

    This function will round the signal and standard error values to 7 decimals places.

    Parameters
    ----------
    df: pd.DataFrame
        Columns: geo_id, timestamp, val, se, sample_size
    export_dir: str
        Export directory
    geo_res: str
        Geographic resolution to which the data has been aggregated
    sensor: str
        Sensor that has been calculated (cumulative_counts vs new_counts)
    metric: Optional[str]
        Metric we are considering, if any.
    start_date: Optional[datetime]
        Earliest date to export or None if no minimum date restrictions should be applied.
    end_date: Optional[datetime]
        Latest date to export or None if no maximum date restrictions should be applied.
    remove_null_samples: Optional[bool]
        Whether to remove entries whose sample sizes are null.
    write_empty_days: Optional[bool]
        If true, every day in between start_date and end_date will have a CSV file written
        even if there is no data for the day. If false, only the days present are written.
    logger: Optional[logging.Logger]
        Pass a logger object here to log information about contradictory missing codes.
    weekly_dates: Optional[bool]
        Whether the output data are weekly or not. If True, will prefix files with
        "weekly_YYYYWW" where WW is the epiweek instead of the usual YYYYMMDD for daily files.

    Returns
    ---------
    dates: pd.Series[datetime]
        Series of dates for which CSV files were exported.
    """
    df = df.copy()

    df["timestamp"] = pd.to_datetime(df["timestamp"])
    if start_date is None:
        start_date = min(df["timestamp"])
    if end_date is None:
        end_date = max(df["timestamp"])
    if not write_empty_days:
        dates = pd.Series(
            df[np.logical_and(df["timestamp"] >= start_date,
                          df["timestamp"] <= end_date)]["timestamp"].unique()
        ).sort_values()
    else:
        dates = pd.date_range(start_date, end_date)

    for date in dates:
        if weekly_dates:
            t = Week.fromdate(pd.to_datetime(str(date)))
            date_str = "weekly_" + str(t.year) + str(t.week).zfill(2)
        else:
            date_str = date.strftime('%Y%m%d')
        if metric is None:
            export_filename = f"{date_str}_{geo_res}_{sensor}.csv"
        else:
            export_filename = f"{date_str}_{geo_res}_{metric}_{sensor}.csv"
        export_file = join(export_dir, export_filename)
        expected_columns = [
            "geo_id",
            "val",
            "se",
            "sample_size",
            "missing_val",
            "missing_se",
            "missing_sample_size"
        ]
        export_df = df[df["timestamp"] == date].filter(items=expected_columns)
        if "missing_val" in export_df.columns:
            export_df = filter_contradicting_missing_codes(
                export_df, sensor, metric, date, logger=logger
            )
        if remove_null_samples:
            export_df = export_df[export_df["sample_size"].notnull()]
        export_df = export_df.round({"val": 7, "se": 7})
        export_df.to_csv(export_file, index=False, na_rep="NA")
    return dates



"""Functions pertaining to running FlaSH daily."""

import numpy as np
import pandas as pd
from scipy.stats import binom
import boto3
from delphi_utils.weekday import Weekday
from .constants import HTML_LINK, STATES
from .. import (
    get_structured_logger,
)


def split_reporting_schedule_dfs(input_df, flash_dir, lag):
    """Separate the input df by reporting schedule (pre-determined).

    Parameters
    ----------
    input_df: the df to split up
    flash_dir: the string reference directory to find all files
    lag: difference between reporting and reference date.

    Returns
    -------
    df of streams updated daily, less often than daily BUT has enough data to use the AR method
    and data that's updated very infrequently and there's not enough data to use the AR method.
    """
    rep_sched = pd.read_csv(f'{flash_dir}/reporting_sched_{lag}.csv', index_col=0)
    min_cut = rep_sched.loc['min_cut'][0]
    rep_sched = rep_sched.drop('min_cut')
    glob_out_list = []
    non_daily_ar = []
    rep_sched.columns = ['schedule']
    for i, df in rep_sched.groupby('schedule'):
        fixed_sum = []
        columns = []
        for col in input_df.columns:
            if col in df.index:
                columns.append(col)
                fixed_sum.append(input_df[col])

        if len(fixed_sum) > 0:
            fixed_sum_df = pd.concat(fixed_sum).to_frame().T
            fixed_sum_df.columns = columns
            fixed_sum_df.index = [input_df.index[0]]
            if i ==1:
                daily_df = fixed_sum_df
            elif i >= min_cut:
                glob_out_list.append(fixed_sum_df)
            else:
                non_daily_ar.append(fixed_sum_df)
    return (daily_df, pd.concat(non_daily_ar,axis=1) , pd.concat(glob_out_list, axis=1))


def bin_approach(df, log=False):
    """Create test statistic.

    Parameters
    ----------
    df with columns of
    y: observed values for streams
    yhat: predicted values for streams
    pop: population for a region

    log: taking the log for the test statistic measure

    Returns
    -------
    today's test-statistic values for the stream
    """
    def ts_dist(x, y, n):
        """Initialize test statistic distribution which is then vectorized."""
        return binom.cdf(x, int(n), y / n)

    vec_ts_dist = np.vectorize(ts_dist)
    if log:
        return pd.DataFrame(vec_ts_dist(np.log(df.y + 2),
                        np.log(df.yhat + 2), np.log(df['pop'] + 2)),
                            index=df.index)
    return pd.DataFrame(vec_ts_dist(df.y, df.yhat, df.pop), index=df.index)




def outlier_detect(df):
    """Determine global outliers by using abs(t-statistic) > 5.

    Parameters
    ----------
    df: Current df to evaluate for global outliers with columns
    for mean and var.

    Returns
    -------
    The columns that are global outliers.
    """
    df.columns = ['x', 'mean', 'var']
    return df.index[((abs(df['x'] - df['mean']) / (df['var'].clip(1))).gt(5))]

def apply_ar(last_7, flash_dir, lag, weekday_correction, non_daily_df, fips_pop_table):
    """Predict y_hat using an AR model.

    Parameters
    ----------
    last_7: the prior 7 days
    flash_dir: the string reference directory to find all files
    lag: the difference between reporing and reference date
    weekday_correction: daily data after weekday correction has been applied
    non_daily_df: df of streams that are not updated daily
    fips_pop_table: df of fips to population

    Returns
    -------
    ts_streams: return of test statistic for the day's steams.
    df_for_ts: dataframe for the test-statistic
    """
    y = pd.concat([weekday_correction, non_daily_df], axis=1)
    y.name = 'y'
    lin_coeff = pd.read_csv(f'{flash_dir}/lin_coeff_{lag}.csv', index_col=0)
    y_hat = pd.Series([np.dot(lin_coeff[x], last_7[x]) for x in y.columns], name='yhat')
    y_hat.index = y.columns
    df_for_ts = y.T.merge(y_hat, left_index=True, right_index=True).merge(fips_pop_table.T
                              , left_index=True, right_index=True)
    df_for_ts.columns = ['y', 'yhat', 'pop']
    ts_streams = bin_approach(df_for_ts, log=True)
    ts_streams.columns = ['test-statistic']
    return ts_streams, df_for_ts


def output(evd_ranking, day, lag, signal, logger):
    """Write the top streams that warrant human inspection to the log.

    Parameters
    ----------
    evd_ranking: the ranking from EVD method (shown to users)
    day: reference date
    lag: difference between reference and report date
    signal: current signal
    logger: logger to write the output of FlaSH

    Returns
    -------
    None
    """
    starter_link = f"{HTML_LINK}{(day+pd.Timedelta(f'{lag}d')).strftime('%Y-%m_%d')}"
    p_text = ""
    for j, (index, value) in enumerate(evd_ranking.sort_values(ascending=False).iteritems()):
        if j < 30:
            start_link = f"{starter_link},{day.strftime('%Y-%m_%d')},{index}"
            p_text += f"\t{start_link}|*{index}*, {'{:.2f}'.format(value)}>\n"
        else:
            break
    name = f"Signal: {signal} Lag: {lag}"
    logger.info(name, payload=p_text)


def evd_ranking_fn(ts_streams, flash_dir):
    """Create ranking using EVDs.

    Parameters
    ----------
    ts_streams: today's test-statistic values for the stream
    flash_dir: the string reference directory to find all files

    Returns
    -------
    evd_ranking: Ranking streams via the extreme value distribution
    """
    EVD_max = pd.read_csv(f'{flash_dir}/max.csv', index_col=0)
    EVD_min = pd.read_csv(f'{flash_dir}/min.csv', index_col=0)
    evd_ranking = pd.concat([ts_streams.apply(lambda x: ts_val(x.values[0],
                             EVD_min['0']), axis=1).sort_values(),
                              ts_streams.apply(lambda x:
                              1 - ts_val(x.values[0], EVD_max['0']),
                              axis=1).sort_values()], axis=1).max(axis=1)
    evd_ranking.name = 'evd_ranking'
    return evd_ranking


def streams_groups_fn(stream, ts_streams):
    """Create the ranking from streams using geographical groupings.

    Uses historical distribution from the test-statistics.

    Parameters
    ----------
    stream: historical test statistic csv
    ts_streams: today's test-statistic values for the stream

    Returns
    -------
    stream_group: the ranking using geographically group test statistic distributions
    """
    streams_groups = stream.copy()
    streams_state = stream[list(filter(lambda x: len(x) == 2,
                                       stream.columns))].unstack().dropna()  # .values
    streams_groups.columns = streams_groups.columns.str[:2]
    ranking_streams = {}
    for key, group in streams_groups.stack().reset_index().groupby('level_1'):
        for col, val in ts_streams.T.iterrows():
            if key == col[:2]:
                total_dist = pd.concat([group[0], streams_state]).reset_index(drop=True)
                ranking_streams[col] = ts_val(val[0], total_dist)
    stream_group = pd.Series(ranking_streams, name='stream_group')
    return stream_group


def setup_fips(flash_dir):
    """Set up fips related dictionaries and population table.

    Input: The directory location for files
    Output: conversion dictionary state to fips & population per fips df
    """
    fips_lookup = pd.read_csv(f'{flash_dir}/fips.csv',
                header=None).astype(str).set_axis(['FIPS', 'STATE'], axis=1)
    fips_lookup['FIPS'] = fips_lookup['FIPS'].str.zfill(2)
    fips_to_STATE = fips_lookup.set_index("FIPS").to_dict()['STATE']
    STATE_to_fips = fips_lookup.set_index("STATE").to_dict()['FIPS']
    fips = pd.read_csv(f'{flash_dir}/geo_fips.csv', index_col=0).set_index('geo')
    orig_fips = fips[fips['pop'].isna()].index
    geos_repl = [fips_to_STATE[x] for x in orig_fips.str[:2]]
    fips.loc[orig_fips, 'pop'] = [float(fips[fips.index == x]['pop']) for x in geos_repl]
    fips_pop_table = fips.unstack().to_frame().T
    fips_pop_table.columns = [STATE_to_fips[x] if x in list(STATES)
                          else x for x in fips_pop_table.columns.droplevel()]
    return STATE_to_fips, fips_pop_table


def ts_val(val, dist):
    """Determine p-value from the test statistic distribution.

    Parameters
    ----------
    val: The test statistic
    dist: The distribution to compare to

    Returns: p-value
    -------

    """
    return sum(val <= dist) / dist.shape[0]

def flash_eval(lag, day, input_df, signal, params, logger=None):
    """Evaluate most recent data using FlaSH.

    Input:
    lag: the difference between the reporting and reference date
    day: the day of the reference date (today is the reporting date)
    input_df: a df from the day for a particular signal that includes natl. state, and county data
    params: additional params needed.
    Ouput:
    None
    """
    if not logger:
        logger = get_structured_logger(
            name=signal,
            filename=params["common"].get("log_filename", None),
            log_exceptions=params["common"].get("log_exceptions", True))

    #TODOv4: Change these to a local dir or aws
    flash_dir = f'flash_ref/{signal}'
    last_7 = pd.read_csv(f'{flash_dir}/last_7_{lag}.csv', index_col=0).astype(float)
    wk_mean = pd.read_csv(f'{flash_dir}/weekday_mean_df_{lag}.csv', index_col=0)
    wk_var = pd.read_csv(f'{flash_dir}/weekday_var_df_{lag}.csv', index_col=0)
    weekday_params = pd.read_csv(f'{flash_dir}/weekday_params_{lag}.csv', index_col=0)
    summary_stats = pd.read_csv(f'{flash_dir}/summary_stats_{lag}.csv', index_col=0)
    summary_stats.index = ['0.25', 'median', '0.75', 'mean', 'var']
    stream = pd.read_csv(f'{flash_dir}/ret_df2_{lag}.csv', index_col=0)


    STATE_to_fips, fips_pop_table = setup_fips(flash_dir)
    input_df.columns = [str(STATE_to_fips[x]) if x in list(STATES)
                        else x for x in input_df.columns]

    # discuss where to do out-of-range handling
    out_range = input_df.columns[input_df.lt(int(params['flash']['support'][0])).iloc[0, :].values
                        | input_df.gt(int(params['flash']['support'][1])).iloc[0, :].values ]

    #only rank streams without out of range data
    input_df = input_df[filter(lambda x: x not in out_range, input_df.columns)]

    daily_update_df, non_daily_df_test, non_ar_df = split_reporting_schedule_dfs(input_df,
                                                                                 flash_dir, lag)

    # only consider non-daily values that are non-0
    non_daily_df_test = non_daily_df_test[non_daily_df_test != 0].dropna(axis=1)

    # Weekday outlier [only for Daily Df]
    weekday_outlier = outlier_detect(daily_update_df.T.merge(wk_mean.loc[day.day_of_week, :],
                      left_index=True, right_index=True).merge(wk_var.loc[day.day_of_week, :],
                      left_index=True, right_index=True))

    # Make weekday correction for daily update
    additive_factor = 1
    weekday_correction = (Weekday.calc_adjustment(
                          weekday_params.loc[daily_update_df.columns, :].to_numpy(), \
                          (daily_update_df + additive_factor).reset_index(),
                           daily_update_df.columns, \
                          'index').set_index('index') - additive_factor).clip(0)

    global_outlier_list = []
    for df in [weekday_correction, non_daily_df_test, non_ar_df]:
        global_outlier_list+=list(outlier_detect(df.T.merge(summary_stats[df.columns].loc['median'
                        ,:],left_index=True, right_index=True
                        ).merge(summary_stats[df.columns].loc['var',:],
                        left_index=True, right_index=True)))

    # Apply AR
    ts_streams, df_for_ts = apply_ar(last_7, flash_dir, lag, weekday_correction,
                        non_daily_df_test, fips_pop_table)
    # find stream ranking (individual)
    stream_individual = ts_streams.T.apply(lambda x:  ts_val(x.values[0],
                                            stream[x.name].dropna()))

    stream_individual.name = 'stream_individual'


    # find stream ranking (group)
    stream_group = streams_groups_fn(stream, ts_streams)

    # find EVD ranking
    evd_ranking = evd_ranking_fn(ts_streams, flash_dir)
    # Save the different categories of outliers/day + rankings for future analysis
    type_of_outlier = pd.DataFrame(index=input_df.columns)
    weekday_frame = weekday_outlier.to_frame()
    weekday_frame[0] = 1
    weekday_frame.columns = ['weekday_outlier']
    type_of_outlier = type_of_outlier.merge(weekday_frame,
            left_index=True, right_index=True, how='outer').fillna(0)
    type_of_outlier['global_outlier'] = 0
    type_of_outlier.loc[global_outlier_list, 'global_outlier'] = 1
    stream_group = stream_group.apply(lambda x: 2 * (0.5 - x) if x < 0.5 else x)
    stream_individual = stream_individual.apply(lambda x: 2 * (0.5 - x) if x < 0.5 else x)
    type_of_outlier = type_of_outlier.merge(stream_individual,
                      left_index=True, right_index=True,
                      how='outer').merge(stream_group,
                      left_index=True, right_index=True, how='outer').merge(evd_ranking,
                      left_index=True, right_index=True, how='outer'
                      ).merge(df_for_ts, left_index=True,
                              right_index=True, how='outer').merge(ts_streams,
                            left_index=True, right_index=True, how='outer')
    type_of_outlier['flash'] = type_of_outlier['evd_ranking']
    indices = type_of_outlier.index[type_of_outlier['evd_ranking'].isna()]
    type_of_outlier.loc[indices, 'flash'] = type_of_outlier.loc[indices, 'global_outlier']
    #if aws parameters are passed, save this dataframe to AWS
    if params.get('archive', None):
        if params['archive'].get("aws_credentials", None):
            session = boto3.Session(
                aws_access_key_id=params['archive']['aws_credentials']["aws_access_key_id"],
                aws_secret_access_key=params['archive']['aws_credentials']["aws_secret_access_key"])
            s3 = session.resource('s3')
            s3.Object(params['flash']["aws_bucket"],
                      f'flags-dev/flash_results/{signal}_{day.strftime("%m_%d_%Y")}_{lag}.csv').put(
                Body=type_of_outlier.to_csv(), ACL='public-read')

    not_fix_daily = list(filter(lambda x: x not in global_outlier_list, daily_update_df.columns))
    not_fix_last_7 = list(filter(lambda x: x not in not_fix_daily, last_7.columns))
    last_7 = pd.concat(
        [pd.concat([last_7[not_fix_daily].iloc[1:, :],
         weekday_correction[not_fix_daily]]).reset_index(drop=True),
         last_7[not_fix_last_7]], axis=1).to_csv(f'{flash_dir}/last_7_{lag}.csv')


    # Save to output log
    output(evd_ranking, day, lag, signal, logger)



"""Utility functions for validation."""
from dataclasses import dataclass, field
from datetime import date, datetime, timedelta
from typing import List
import pandas as pd

# Recognized geo types.
GEO_REGEX_DICT = {
    'county': r'^\d{5}$',
    'hrr': r'^\d{1,3}$',
    'hhs': r'^\d{1,2}$',
    'msa': r'^\d{5}$',
    'dma': r'^\d{3}$',
    'state': r'^[a-zA-Z]{2}$',
    'nation': r'^[a-zA-Z]{2}$'
}


@dataclass
class TimeWindow:
    """Object to store a window of time ending on `end_date` and lasting `span_length`."""

    end_date: date
    span_length: timedelta
    start_date: date = field(init=False)
    date_seq: List[date] = field(init=False)

    def __post_init__(self):
        """Derive the start date of this span."""
        self.start_date = self.end_date - self.span_length + timedelta(days = 1)
        self.date_seq = [self.start_date + timedelta(days=x)
                         for x in range(self.span_length.days)]

    @classmethod
    def from_params(cls, end_date_str: str, span_length_int: int):
        """Create a TimeWindow from param representations of its members."""
        span_length = timedelta(days=span_length_int)
        if end_date_str.startswith("today"):
            if end_date_str == "today":
                days_back = 0
            else:
                assert end_date_str.startswith("today-")
                days_back = int(end_date_str[6:])
            end_date = date.today() - timedelta(days=days_back)
        elif end_date_str.startswith("sunday+"):
            days_back = (date.isoweekday() - int(end_date_str[7:])) % 7
            end_date = date.today() - timedelta(days=days_back)
        else:
            end_date = datetime.strptime(end_date_str, '%Y-%m-%d').date()
        return cls(end_date, span_length)


def relative_difference_by_min(x, y):
    """Calculate relative difference between two numbers."""
    return (x - y) / min(x, y)


def aggregate_frames(frames_list):
    """Aggregate a list of data frames into a single frame.

    Parameters
    ----------
    frames_list: List[Tuple(str, re.match, pd.DataFrame)]
        triples of filenames, filename matches with the geo regex, and the data from the file

    Returns
    -------
    A pd.DataFrame concatenation of all data frames in `frames_list` with additional columns for
    geo_type, time_value, and signal derived from the corresponding re.match.
    """
    all_frames = []
    for _, match, data_df in frames_list:
        df = data_df.copy()
        # Get geo_type, date, and signal name as specified by CSV name.
        df['geo_type'] = match.groupdict()['geo_type']
        df['time_value'] = datetime.strptime(
            match.groupdict()['date'], "%Y%m%d").date()
        df['signal'] = match.groupdict()['signal']

        all_frames.append(df)

    return pd.concat(all_frames).reset_index(drop=True)

def lag_converter(lag_dict):
    """Convert a dictionary of lag values into the proper format.

    Parameters
    ----------
    lag_dict: Dict[str, str or int]
        Keys are either 'all', or signal names
        Values are either numeric or days of the week, represented by
            "Sunday+[0-7],[0-9]", first part represents the day of the week
            the upload happens, while the second number represents lag upon
            upload.
    Returns
    ----------
    Dict[str, int]
        Keys are all active signal names for an indicator, or "all"
        Values are obtained by looking at signal name entries,
            If not present then by the 'all' indicator,
            else 1.
    """
    def value_interpret(value):
        """Convert value from string to numeric, including sunday+m,n."""
        if isinstance(value, int):
            value_num = value
        elif value.startswith("sunday+"):
            # Check that the number proceeding sunday+ has length of 1
            assert value[8] == ","
            # Value_num calculates the number of days between today and the
            # Update day, if both days of the week are the same,
            # Expected lag is 0
            value_num = (date.today().isoweekday() - int(value[7:8])) % 7
            # Add on expected lag to lag from weekdays
            value_num += int(value[9:])
        else:
            value_num = int(value)
        return value_num

    # Converting strings to numeric output
    output_dict = {sig:value_interpret(lag_dict.get(
        sig)) for sig in lag_dict.keys()}
    return output_dict

def end_date_helper(params):
    """Calculate end_date from params."""
    common_params = params["common"]
    min_expected_lag = lag_converter(common_params.get(
            "min_expected_lag", dict()))
    if len(min_expected_lag) == 0:
        min_expected_lag_overall = 1
    else:
        min_expected_lag_overall = min(min_expected_lag.values())
    end_date = common_params.get(
        "end_date", f"today-{min_expected_lag_overall}")
    return end_date



"""Smoother utility.

This file contains the smoothing utility functions. We have a number of
possible smoothers to choose from: windowed average, local weighted regression,
and a causal Savitzky-Golay filter.

Code is courtesy of Dmitry Shemetov, Maria Jahja, and Addison Hu.

These smoothers are all functions that take a 1D numpy array and return a smoothed
1D numpy array of the same length (with a few np.nans in the beginning). See the
docstrings for details.
"""

from typing import Union
import warnings

import numpy as np
import pandas as pd


class Smoother:  # pylint: disable=too-many-instance-attributes
    """Smoother class.

    This is the smoothing utility class. This class holds the parameter settings for its smoother
    methods and provides reasonable defaults. Basic usage can be found in the examples below.

    The smoother function takes numpy arrays or pandas Series as input, expecting the values to be
    on a regularly-spaced time grid. NANs are ok, as long as the array does not begin with a NAN.
    The rest of the NANs will be handled via imputation by default, though this can be turned off.

    Parameters
    ----------
    smoother_name: {'savgol', 'moving_average', 'identity', 'left_gauss_linear'}
        This variable specifies the smoother. We have four smoothers, currently:
        * 'savgol' or a Savtizky-Golay smoother (default)
        * 'moving_average' or a moving window average smoother
        * 'identity' or the trivial smoother (no smoothing)
        * 'left_gauss_linear' or a Gaussian-weight linear regression smoother
        Descriptions of the smoothers are available in the doc strings. Full mathematical
        details are in: https://github.com/cmu-delphi/covidcast-modeling/ in the folder
        'indicator_smoother'.
    poly_fit_degree: int
        A parameter for the 'savgol' smoother which sets the degree of the polynomial fit.
    window_length: int
        The length of the fitting window for 'savgol' and the averaging window 'moving_average'.
        This value is in the units provided by the data, which are likely to be days for Delphi.
        Note that if window_length is smaller than the length of the signal, then only the
        imputation method is run on the signal.
    gaussian_bandwidth: float or None
        If float, all regression is done with Gaussian weights whose variance is
        half the gaussian_bandwidth. If None, performs unweighted regression. (Applies
        to 'left_gauss_linear' and 'savgol'.)
        Here are some reference values (the given bandwidth produces a 95% weighting on
        the data of length time window into the past):
        time window    |     bandwidth
        7                        36
        14                       144
        21                       325
        28                       579
        35                       905
        42                       1303
    impute: {'savgol', 'zeros', None}
        If 'savgol' (default), will fill nan values with a savgol fit on the largest available time
        window prior (up to window_length). If 'zeros', will fill nan values with zeros.
        If None, leaves the nans in place.
    minval: float or None
        The smallest value to allow in a signal. If None, there is no smallest value.
        Currently only implemented for 'left_gauss_linear'. This should probably not be in the scope
        of the smoothing utility.
    boundary_method: {'shortened_window', 'identity', 'nan'}
        Determines how the 'savgol' method handles smoothing at the (left) boundary, where the past
        data length is shorter than the window_length parameter. If 'shortened_window', it uses the
        maximum window available; at the very edge (generally up to poly_fit_degree) it keeps the
        same value as the raw signal. If 'identity', it just keeps the raw signal. If 'nan', it
        writes nans. For the other smoothing methods, 'moving_average' writes nans and
        'left_gauss_linear' uses a shortened window.

    Methods
    ----------
    smooth: np.ndarray or pd.Series
        Takes a 1D signal and returns a smoothed version.
        The input and the output have the same length and type.

    Example Usage
    -------------
    Example 1. Apply a rolling average smoother with a window of length 10.
    >>> smoother = Smoother(smoother_name='moving_average', window_length=10)
    >>> smoothed_signal = smoother.smooth(signal)

    Example 2. Smooth a dataframe column.
    >>> smoother = Smoother(smoother_name='savgol')
    >>> df[col] = df[col].transform(smoother.smooth)

    Example 3. Apply a rolling weighted average smoother, with 95% weight on the recent 2 weeks and
               a sharp cutoff after 4 weeks.
    >>> smoother = Smoother(smoother_name='savgol', poly_fit_degree=0, window_length=28,
                          gaussian_bandwidth=144)
    >>> smoothed_signal = smoother.smooth(signal)

    Example 4. Apply a local linear regression smoother (essentially equivalent to
               `left_gauss_linear`), with 95% weight on the recent week and a sharp
               cutoff after 3 weeks.
    >>> smoother = Smoother(smoother_name='savgol', poly_fit_degree=1, window_length=21,
                          gaussian_bandwidth=36)
    >>> smoothed_signal = smoother.smooth(signal)

    Example 5. Apply the identity function (simplifies code that iterates through smoothers _and_
               expects a copy of the raw data).
    >>> smoother = Smoother(smoother_name='identity')
    >>> smoothed_signal = smoother.smooth(signal)
    """

    def __init__(
        self,
        smoother_name="savgol",
        poly_fit_degree=2,
        window_length=28,
        gaussian_bandwidth=144,  # a ~2 week window
        impute_method=None,
        minval=None,
        boundary_method="shortened_window",
    ):
        """See class docstring."""
        self.smoother_name = smoother_name
        self.poly_fit_degree = poly_fit_degree
        self.window_length = window_length
        self.gaussian_bandwidth = gaussian_bandwidth
        self.impute_method = self._select_imputer(impute_method, self.smoother_name)
        self.minval = minval
        self.boundary_method = boundary_method

        valid_smoothers = {"savgol", "left_gauss_linear", "moving_average", "identity"}
        valid_impute_methods = {"savgol", "zeros", "identity"}
        valid_boundary_methods = {"shortened_window", "identity", "nan"}
        if self.smoother_name not in valid_smoothers:
            raise ValueError("Invalid smoother_name given.")
        if self.impute_method not in valid_impute_methods:
            raise ValueError("Invalid impute_method given.")
        if self.boundary_method not in valid_boundary_methods:
            raise ValueError("Invalid boundary_method given.")
        if self.window_length <= 1:
            raise ValueError("Window length is too short.")

        if smoother_name == "savgol":
            # The polynomial fitting is done on a past window of size window_length
            # including the current day value.
            self.coeffs = self.savgol_coeffs(
                -self.window_length + 1, 0, self.poly_fit_degree
            )
        else:
            self.coeffs = None

    def _select_imputer(self, impute_method, smoother_name):
        if impute_method is None and smoother_name != "identity":
            return "savgol"
        if impute_method is None and smoother_name == "identity":
            return "identity"
        return impute_method

    def smooth(
        self, signal: Union[np.ndarray, pd.Series], impute_order=2
    ) -> Union[np.ndarray, pd.Series]:
        """Apply a smoother to a signal.

        The major workhorse smoothing function. Imputes the nans and then applies
        a smoother to the signal.

        Parameters
        ----------
        signal: np.ndarray or pd.Series
            A 1D signal to be smoothed.
        impute_order: int
            The polynomial order of the fit used for imputation. By default, this is set to
            2.

        Returns
        ----------
        signal_smoothed: np.ndarray or pd.Series
            A smoothed 1D signal. Returns an array of the same type and length as
            the input.
        """
        # If all nans, pass through
        if np.all(np.isnan(signal)):
            return signal

        is_pandas_series = isinstance(signal, pd.Series)
        pandas_index = signal.index if is_pandas_series else None
        signal = signal.to_numpy() if is_pandas_series else signal

        # Find where the first non-nan value is located and truncate the initial nans
        ix = np.where(~np.isnan(signal))[0][0]
        signal = signal[ix:]

        # Don't smooth in certain edge cases
        if len(signal) < self.poly_fit_degree or len(signal) == 1:
            signal_smoothed = signal.copy()
        else:
            # Impute
            signal = self.impute(signal, impute_order=impute_order)

            # Smooth
            signal_smoothed = self._select_smoother()(signal)

        # Append the nans back, since we want to preserve length
        signal_smoothed = np.hstack([np.nan * np.ones(ix), signal_smoothed])
        # Convert back to pandas if necessary
        if is_pandas_series:
            signal_smoothed = pd.Series(signal_smoothed)
            signal_smoothed.index = pandas_index
        return signal_smoothed

    def _select_smoother(self):
        """Select a smoothing method based on the smoother type."""
        if self.smoother_name == "savgol":
            return self.savgol_smoother
        if self.smoother_name == "left_gauss_linear":
            return self.left_gauss_linear_smoother
        if self.smoother_name == "moving_average":
            return self.moving_average_smoother
        if self.smoother_name == "identity":
            return lambda x: x
        raise ValueError(f"invalid smoother {self.smoother_name}")

    def impute(self, signal, impute_order=2):
        """Impute the nan values in the signal.

        See the class docstring for an explanation of the impute methods.

        Parameters
        ----------
        signal: np.ndarray
            1D signal to be imputed.
        impute_order: int
            The polynomial order of the fit used for imputation.

        Returns
        -------
        imputed_signal: np.ndarray
            Imputed signal.
        """
        if self.impute_method == "savgol":
            # We cannot impute if the signal begins with a NaN (there is no information to go by).
            # To preserve input-output array lengths, this util will not drop NaNs for you.
            if np.isnan(signal[0]):
                raise ValueError("The signal should not begin with a nan value.")
            imputed_signal = self.savgol_impute(signal, impute_order)
        elif self.impute_method == "zeros":
            imputed_signal = np.nan_to_num(signal)
        elif self.impute_method == "identity":
            imputed_signal = np.copy(signal)

        return imputed_signal

    def moving_average_smoother(self, signal):
        """Compute a moving average on the signal.

        Parameters
        ----------
        signal: np.ndarray
            Input array.

        Returns
        -------
        signal_smoothed: np.ndarray
            An array with the same length as arr, but the first window_length-1
            entries are np.nan.
        """
        if not isinstance(self.window_length, int):
            raise ValueError("k must be int.")

        signal_padded = np.append(np.nan * np.ones(self.window_length - 1), signal)
        signal_smoothed = (
            np.convolve(
                signal_padded, np.ones(self.window_length, dtype=int), mode="valid"
            )
            / self.window_length
        )

        return signal_smoothed

    def left_gauss_linear_smoother(self, signal):
        """Smooth the y-values using a local linear regression with Gaussian weights.

        DEPRECATED: This method is available to help sanity check the 'savgol' method.
        Use 'savgol' with poly_fit_degree=1 and the appropriate gaussian_bandwidth instead.

        At each time t, we use the data from times 1, ..., t-dt, weighted
        using the Gaussian kernel, to produce the estimate at time t.

        Parameters
        ----------
        signal: np.ndarray
            A 1D signal.

        Returns
        ----------
        signal_smoothed: np.ndarray
            A smoothed 1D signal.
        """
        warnings.warn(
            "Use the savgol smoother with poly_fit_degree=1 instead.",
            DeprecationWarning,
        )
        n = len(signal)
        signal_smoothed = np.zeros_like(signal)
        # A is the regression design matrix
        A = np.vstack([np.ones(n), np.arange(n)]).T  # pylint: disable=invalid-name
        for idx in range(n):
            weights = np.exp(
                -((np.arange(idx + 1) - idx) ** 2) / self.gaussian_bandwidth
            )
            AwA = np.dot(  # pylint: disable=invalid-name
                A[: (idx + 1), :].T * weights, A[: (idx + 1), :]
            )
            Awy = np.dot(  # pylint: disable=invalid-name
                A[: (idx + 1), :].T * weights, signal[: (idx + 1)].reshape(-1, 1)
            )
            try:
                beta = np.linalg.solve(AwA, Awy)
                signal_smoothed[idx] = np.dot(A[: (idx + 1), :], beta)[-1]
            except np.linalg.LinAlgError:
                signal_smoothed[idx] = (
                    signal[idx]  # pylint: disable=using-constant-test
                    if self.impute
                    else np.nan
                )
        if self.minval is not None:
            signal_smoothed[signal_smoothed <= self.minval] = self.minval
        return signal_smoothed

    def savgol_predict(self, signal, poly_fit_degree, nr):
        """Predict a single value using the savgol method.

        Fits a polynomial through the values given by the signal and returns the value
        of the polynomial at the right-most signal-value. More precisely, for a signal of length
        n, fits a poly_fit_degree polynomial through the points signal[-n+1+nr], signal[-n+2+nr],
        ..., signal[nr], and returns the evaluation of the polynomial at signal[0]. Hence, if
        nr=0, then the last value of the signal is smoothed, and if nr=-1, then the value after
        the last signal value is anticipated.

        Parameters
        ----------
        signal: np.ndarray
            A 1D signal to smooth.
        poly_fit_degree: int
            The degree of the polynomial fit.
        nr: int
            An integer that determines the position of the predicted value relative to the signal.

        Returns
        ----------
        predicted_value: float
            The anticipated value that comes after the end of the signal based on a polynomial fit.
        """
        coeffs = self.savgol_coeffs(-len(signal) + 1 + nr, nr, poly_fit_degree)
        predicted_value = signal @ coeffs
        return predicted_value

    def savgol_coeffs(self, nl, nr, poly_fit_degree):
        """Solve for the Savitzky-Golay coefficients.

        Solves for the Savitzky-Golay coefficients. The coefficients c_i
        give a filter so that
            y = sum_{i=-{n_l}}^{n_r} c_i x_i
        is the value at 0 (thus the constant term) of the polynomial fit
        through the points {x_i}. The coefficients are c_i are calculated as
            c_i =  ((A.T @ A)^(-1) @ (A.T @ e_i))_0
        where A is the design matrix of the polynomial fit and e_i is the standard
        basis vector i. This is currently done via a full inversion, which can be
        optimized.

        Parameters
        ----------
        nl: int
            The left window bound for the polynomial fit, inclusive.
        nr: int
            The right window bound for the polynomial fit, inclusive.
        poly_fit_degree: int
            The degree of the polynomial to be fit.

        Returns
        ----------
        coeffs: np.ndarray
            A vector of coefficients of length nr - nl + 1 that determines the savgol
            convolution filter.
        """
        if nl >= nr:
            raise ValueError("The left window bound should be less than the right.")
        if nr > 0:
            warnings.warn("The filter is no longer causal.")

        A = np.vstack(  # pylint: disable=invalid-name
            [np.arange(nl, nr + 1) ** j for j in range(poly_fit_degree + 1)]
        ).T

        if self.gaussian_bandwidth is None:
            mat_inverse = np.linalg.inv(A.T @ A) @ A.T
        else:
            weights = np.exp(-((np.arange(nl, nr + 1)) ** 2) / self.gaussian_bandwidth)
            mat_inverse = np.linalg.inv((A.T * weights) @ A) @ (A.T * weights)
        window_length = nr - nl + 1
        coeffs = np.zeros(window_length)
        for i in range(window_length):
            basis_vector = np.zeros(window_length)
            basis_vector[i] = 1.0
            coeffs[i] = (mat_inverse @ basis_vector)[0]
        return coeffs

    def savgol_smoother(self, signal):  # pylint: disable=inconsistent-return-statements
        """Smooth signal with the savgol smoother.

        Returns a convolution of the 1D signal with the Savitzky-Golay coefficients, respecting
        boundary effects. For an explanation of boundary effects methods, see the class docstring.

        Parameters
        ----------
        signal: np.ndarray
            A 1D signal.

        Returns
        ----------
        signal_smoothed: np.ndarray
            A smoothed 1D signal of same length as signal.
        """
        # Reverse because np.convolve reverses the second argument
        temp_reversed_coeffs = np.array(list(reversed(self.coeffs)))

        # Smooth the part of the signal away from the boundary first
        signal_padded = np.append(np.nan * np.ones(len(self.coeffs) - 1), signal)
        signal_smoothed = np.convolve(signal_padded, temp_reversed_coeffs, mode="valid")

        # This section handles the smoothing behavior at the (left) boundary:
        # - shortened_window (default) applies savgol with a smaller window to do the fit
        # - identity keeps the original signal (doesn't smooth)
        # - nan writes nans
        if self.boundary_method == "nan":
            return signal_smoothed

        # boundary methods "identity" and "shortened window"
        for ix in range(min(len(self.coeffs), len(signal))):
            if ix == 0 or self.boundary_method == "identity":
                signal_smoothed[ix] = signal[ix]
            else:
                # At the very edge, the design matrix is often singular, in which case
                # we just fall back to the raw signal
                try:
                    signal_smoothed[ix] = self.savgol_predict(
                        signal[: ix + 1], self.poly_fit_degree, 0
                    )
                except np.linalg.LinAlgError:  # for small ix, the design matrix is singular
                    signal_smoothed[ix] = signal[ix]
        return signal_smoothed

    def savgol_impute(self, signal, impute_order):
        """Impute the nan values in signal using savgol.

        This method fills the nan values in the signal with polynomial interpolation
        on a rolling window of the immediate past up to window_length data points.

        A number of boundary cases are handled involving nan filling close to the boundary.

        Note that in the case of many adjacent nans, the method will use previously
        imputed values to do the fitting for later values.

        Parameters
        ----------
        signal: np.ndarray
            A 1D signal to be imputed.
        impute_order: int
            The polynomial order of the fit used for imputation.

        Returns
        ----------
        signal_imputed: np.ndarray
            An imputed 1D signal.
        """
        if impute_order > self.window_length:
            raise ValueError("Impute order must be smaller than window length.")

        signal_imputed = np.copy(signal)
        for ix in np.where(np.isnan(signal_imputed))[0]:
            # Boundary cases
            if ix < self.window_length:
                # At the boundary, a single value should just be extended
                if ix == 1:
                    signal_imputed[ix] = signal_imputed[ix - 1]
                # Otherwise, use savgol fitting on the largest window prior,
                # reduce the polynomial degree if needed (can't fit if the
                # imputation order is larger than the available data)
                else:
                    signal_imputed[ix] = self.savgol_predict(
                        signal_imputed[:ix], min(ix - 1, impute_order), -1
                    )
            # Away from the boundary, use savgol fitting on a fixed window
            else:
                signal_imputed[ix] = self.savgol_predict(
                    signal_imputed[ix - self.window_length : ix],
                    impute_order,
                    -1,
                )
        return signal_imputed



# -*- coding: utf-8 -*-
"""Functions to call when running the tool.

This module should contain a function called `run_module`, that is executed
when the module is run with `python -m delphi_utils.flash_eval`.
"""
from datetime import date
import pandas as pd
from .eval_day import flash_eval
from ..validator.datafetcher import read_filenames, load_csv

def run_module(params):
    """Run the FlaSH module.

    The parameters dictionary must include the signals and signals.
    We are only considering lag-1 data.
    """
    if params.get("flash", None):
        signals = params["flash"].get("signals", [])
        for signal in signals:
            export_files = read_filenames(params["common"]["export_dir"])
            days = {}
            for (x, _) in export_files:
                #Concat the data from recent files at nation, state, and county resolution per day.
                if signal in x and pd.Series([y in x for y in ['state', 'county', 'nation']]).any():
                    day = pd.to_datetime(x.split('_')[0], format="%Y%m%d", errors='raise')
                    days[day] = pd.concat([days.get(day, pd.DataFrame()),
                                           load_csv(f"{params['common']['export_dir']}/{x}")])

            for day, input_df in dict(sorted(days.items())).items():
                input_df = input_df[['geo_id', 'val']].set_index('geo_id').T
                input_df.index = [day]
                today = date.today()
                lag= (pd.to_datetime(today)-pd.to_datetime(day)).days
                # inital flash implementation assume lag == 1 always
                #if str(lag) in params["flash"]["lags"]:
                lag=1
                flash_eval(int(lag), day, input_df, signal, params)



"""Constants used in FlaSH."""

#Regions considered under states
STATES = ['ak', 'al', 'ar', 'as', 'az', 'ca', 'co', 'ct', 'dc', 'de', 'fl', 'ga',
            'gu', 'hi', 'ia', 'id', 'il', 'in', 'ks', 'ky', 'la',
            'ma', 'md', 'me', 'mi', 'mn', 'mo', 'mp', 'ms', 'mt', 'nc',
            'nd', 'ne', 'nh', 'nj', 'nm', 'nv', 'ny', 'oh', 'ok',
            'or', 'pa', 'pr', 'ri', 'sc', 'sd', 'tn', 'tx', 'ut', 'va', 'vi', 'vt',
            'wa', 'wi', 'wv', 'wy']

#HTML Link for the visualization tool alerts
HTML_LINK = "<https://ananya-joshi-visapp-vis-523f3g.streamlitapp.com/?params="



"""Call the function run_module when executed.

This file indicates that calling the module (`python -m MODULE_NAME`) will
call the function `run_module` found within the run.py file. There should be
no need to change this template.
"""

from delphi_utils import read_params
from .run import run_module  # pragma: no cover

run_module(read_params())  # pragma: no cover



"""Static file checks."""
import re
from datetime import datetime
from dataclasses import dataclass
from typing import Dict, List
import pandas as pd
from .datafetcher import FILENAME_REGEX
from .errors import ValidationFailure
from .utils import GEO_REGEX_DICT, TimeWindow, lag_converter
from ..geomap import GeoMapper

class StaticValidator:
    """Class for validation of static properties of individual datasets."""

    @dataclass
    class Parameters:
        """Configuration parameters."""

        # Span of time over which to perform checks
        time_window: TimeWindow
        # Threshold for reporting small sample sizes
        minimum_sample_size: int
        # Whether to report missing standard errors
        missing_se_allowed: bool
        # Whether to report missing sample sizes
        missing_sample_size_allowed: bool
        # Valid geo values not found in the GeoMapper
        additional_valid_geo_values: Dict[str, List[str]]
        # how many days behind do we expect each signal to be
        max_expected_lag: Dict[str, int]

    def __init__(self, params):
        """
        Initialize object and set parameters.

        Arguments:
            - params: dictionary of user settings; if empty, defaults will be used
        """
        common_params = params["common"]
        static_params = params.get("static", dict())

        self.params = self.Parameters(
            time_window = TimeWindow.from_params(common_params["end_date"],
                                                 common_params["span_length"]),
            minimum_sample_size = static_params.get('minimum_sample_size', 100),
            missing_se_allowed = static_params.get('missing_se_allowed', False),
            missing_sample_size_allowed = static_params.get('missing_sample_size_allowed', False),
            additional_valid_geo_values = static_params.get('additional_valid_geo_values', {}),
            max_expected_lag=lag_converter(common_params.get("max_expected_lag", dict()))
        )


    def validate(self, file_list, report):
        """
        Perform checks over single-file data sets.

        Parameters
        ----------
        loaded_data: List[Tuple(str, re.match, pd.DataFrame)]
            triples of filenames, filename matches with the geo regex, and the data from the file
        report: ValidationReport
            report to which the results of these checks will be added
        """
        self.check_missing_date_files(file_list, report)

        # Individual file checks
        # For every daily file, read in and do some basic format and value checks.
        for filename, match, data_df in file_list:
            self.check_df_format(data_df, filename, report)
            self.check_duplicate_rows(data_df, filename, report)
            self.check_bad_geo_id_format(
                data_df, filename, match.groupdict()['geo_type'], report)
            self.check_bad_geo_id_value(
                data_df, filename, match.groupdict()['geo_type'], report)
            self.check_bad_val(data_df, filename, match.groupdict()['signal'], report)
            self.check_bad_se(data_df, filename, report)
            self.check_bad_sample_size(data_df, filename, report)


    def check_missing_date_files(self, daily_filenames, report):
        """
        Check for missing dates between the specified start and end dates.

        Arguments:
            - daily_filenames: List[Tuple(str, re.match, pd.DataFrame)]
                triples of filenames, filename matches with the geo regex, and the data from the
                file
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        # Check to see if there are any files in the export directory
        # Validator will throw an error if the directory is empty, which can be suppressed
        if len(daily_filenames) == 0:
            report.add_raised_error(
                ValidationFailure("check_empty_filelist",
                                  message="No files found in export directory"))
        # Check for missing date only happens when files are found
        else:
            # Create set of all dates seen in CSV names.
            unique_dates = {datetime.strptime(
                daily_filename[0][0:8], '%Y%m%d').date() for daily_filename in daily_filenames}
            # Diff expected and observed dates.
            expected_dates = self.params.time_window.date_seq
            if len(self.params.max_expected_lag) == 0:
                max_expected_lag_overall = 10
            else:
                max_expected_lag_overall = max(self.params.max_expected_lag.values())

            # Only check for date if it should definitely be present,
            # i.e if it is more than max_expected_lag since the checking date
            expected_dates = [date for date in expected_dates if
                ((datetime.today().date() - date).days) > max_expected_lag_overall]
            check_dateholes = list(set(expected_dates).difference(unique_dates))
            check_dateholes.sort()

            if check_dateholes:
                report.add_raised_error(
                    ValidationFailure("check_missing_date_files",
                                  message="Missing dates are observed; if these dates are already "
                                          "in the API they would not be updated"))

        report.increment_total_checks()

    def check_df_format(self, df_to_test, nameformat, report):
        """
        Check basic format of source data CSV df.

        Arguments:
            - df_to_test: pandas dataframe of a single CSV of source data
            (one day-signal-geo_type combo)
            - nameformat: str CSV name; for example, "20200624_county_smoothed_nohh_cmnty_cli.csv"
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        pattern_found = FILENAME_REGEX.match(nameformat)
        if not nameformat or not pattern_found:
            report.add_raised_error(
                ValidationFailure("check_filename_format",
                                  filename=nameformat,
                                  message="nameformat not recognized"))

        report.increment_total_checks()

        if not isinstance(df_to_test, pd.DataFrame):
            report.add_raised_error(
                ValidationFailure(
                    "check_file_data_format",
                    filename=nameformat,
                    message=f"expected pd.DataFrame but got {type(df_to_test)}."))

        report.increment_total_checks()

    def _get_valid_geo_values(self, geo_type):
        # geomapper uses slightly different naming conventions for geo_types
        if geo_type == "state":
            geomap_type = "state_id"
        elif geo_type == "county":
            geomap_type = "fips"
        else:
            geomap_type = geo_type

        gmpr = GeoMapper()
        valid_geos = gmpr.get_geo_values(geomap_type)
        valid_geos |= set(self.params.additional_valid_geo_values.get(geo_type, []))
        if geo_type == "county":
            valid_geos |= set(x + "000" for x in gmpr.get_geo_values("state_code"))
        return valid_geos

    def check_bad_geo_id_value(self, df_to_test, filename, geo_type, report):
        """
        Check for bad geo_id values, by comparing to a list of known historical values.

        Arguments:
            - df_to_test: pandas dataframe of CSV source data containing the geo_id column to check
            - geo_type: string from CSV name specifying geo type (state, county, msa, etc.) of data
            - report: ValidationReport; report where results are added
        """
        valid_geos = self._get_valid_geo_values(geo_type)
        unexpected_geos = [geo for geo in df_to_test['geo_id']
                           if geo.lower() not in valid_geos]
        if len(unexpected_geos) > 0:
            report.add_raised_error(
                ValidationFailure(
                    "check_bad_geo_id_value",
                    filename=filename,
                    message=f"Unrecognized geo_ids (not in historical data) {unexpected_geos}"))
        report.increment_total_checks()
        upper_case_geos = [
            geo for geo in df_to_test['geo_id'] if geo.lower() != geo]
        if len(upper_case_geos) > 0:
            report.add_raised_warning(
                ValidationFailure(
                    "check_geo_id_lowercase",
                    filename=filename,
                    message=f"geo_ids {upper_case_geos} contains uppercase characters. Lowercase "
                            "is preferred."))
        report.increment_total_checks()

    def check_bad_geo_id_format(self, df_to_test, nameformat, geo_type, report):
        """
        Check validity of geo_type and format of geo_ids, according to regex pattern.

        Arguments:
            - df_to_test: pandas dataframe of CSV source data
            - geo_type: string from CSV name specifying geo type (state, county, msa, hrr) of data
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        def find_all_unexpected_geo_ids(df_to_test, geo_regex, geo_type):
            """Check if any geo_ids in df_to_test aren't formatted correctly wrt geo_regex."""
            numeric_geo_types = {"msa", "county", "hrr", "dma"}
            fill_len = {"msa": 5, "county": 5, "dma": 3}

            if geo_type in numeric_geo_types:
                # Check if geo_ids were stored as floats (contain decimal point) and
                # contents before decimal match the specified regex pattern.
                leftover = [geo[1] for geo in df_to_test["geo_id"].str.split(
                    ".") if len(geo) > 1 and re.match(geo_regex, geo[0])]

                # If any floats found, remove decimal and anything after.
                if len(leftover) > 0:
                    df_to_test["geo_id"] = [geo[0]
                                            for geo in df_to_test["geo_id"].str.split(".")]

                    report.add_raised_warning(
                        ValidationFailure(
                            "check_geo_id_type",
                            filename=nameformat,
                            message="geo_ids saved as floats; strings preferred"))

            if geo_type in fill_len.keys():
                # Left-pad with zeroes up to expected length. Fixes missing leading zeroes
                # caused by FIPS codes saved as numeric.
                df_to_test["geo_id"] = pd.Series([geo.zfill(fill_len[geo_type])
                                                  for geo in df_to_test["geo_id"]], dtype=str)

            expected_geos = [geo[0] for geo in df_to_test['geo_id'].str.findall(
                geo_regex) if len(geo) > 0]

            unexpected_geos = {geo for geo in set(
                df_to_test['geo_id']) if geo not in expected_geos}

            if len(unexpected_geos) > 0:
                report.add_raised_error(
                    ValidationFailure(
                        "check_geo_id_format",
                        filename=nameformat,
                        message=f"Non-conforming geo_ids {unexpected_geos} found"))

        if geo_type in GEO_REGEX_DICT:
            find_all_unexpected_geo_ids(
                df_to_test, GEO_REGEX_DICT[geo_type], geo_type)
        else:
            report.add_raised_error(
                ValidationFailure(
                    "check_geo_type",
                    filename=nameformat,
                    message=f"Unrecognized geo type {geo_type}"))

        report.increment_total_checks()

    def check_bad_val(self, df_to_test, nameformat, signal_type, report):
        """
        Check value field for validity.

        Arguments:
            - df_to_test: pandas dataframe of a single CSV of source data
            - signal_type: string from CSV name specifying signal type (smoothed_cli, etc) of data
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        # Determine if signal is a proportion (# of x out of 100k people) or percent
        percent_option = bool('pct' in signal_type)
        proportion_option = bool('prop' in signal_type)

        if percent_option:
            if not df_to_test[(df_to_test['val'] > 100)].empty:
                report.add_raised_error(
                    ValidationFailure(
                        "check_val_pct_gt_100",
                        filename=nameformat,
                        message="val column can't have any cell greater than 100 for percents"))

            report.increment_total_checks()

        if proportion_option:
            if not df_to_test[(df_to_test['val'] > 100000)].empty:
                report.add_raised_error(
                    ValidationFailure("check_val_prop_gt_100k",
                                      filename=nameformat,
                                      message="val column can't have any cell greater than 100000 "
                                              "for proportions"))

            report.increment_total_checks()

        if not df_to_test[(df_to_test['val'] < 0)].empty:
            report.add_raised_error(
                ValidationFailure("check_val_lt_0",
                                  filename=nameformat,
                                  message="val column can't have any cell smaller than 0"))

        report.increment_total_checks()

    def check_bad_se(self, df_to_test, nameformat, report):
        """
        Check standard errors for validity.

        Arguments:
            - df_to_test: pandas dataframe of a single CSV of source data
            (one day-signal-geo_type combo)
            - nameformat: str CSV name; for example, "20200624_county_smoothed_nohh_cmnty_cli.csv"
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        if self.params.missing_se_allowed:
            result = df_to_test.query(
                '~(se.isnull() | (se >= 0))')

            if not result.empty:
                report.add_raised_error(
                    ValidationFailure("check_se_missing_or_in_range",
                                      filename=nameformat,
                                      message="se must be NA or non-negative"))

            report.increment_total_checks()
        else:
            # Find rows not in the allowed range for se.
            result = df_to_test.query(
                '~(se >= 0)')

            if not result.empty:
                report.add_raised_error(
                    ValidationFailure("check_se_not_missing_and_in_range",
                                      filename=nameformat,
                                      message="se must be non-negative and not "
                                              "missing"))

            report.increment_total_checks()

            if df_to_test["se"].isnull().mean() > 0.5:
                report.add_raised_error(
                    ValidationFailure("check_se_many_missing",
                                      filename=nameformat,
                                      message='Recent se values are >50% NA'))

            report.increment_total_checks()

        result_jeffreys = df_to_test.query('(val == 0) & (se == 0)')
        result_alt = df_to_test.query('se == 0')

        if not result_jeffreys.empty:
            report.add_raised_error(
                ValidationFailure("check_se_0_when_val_0",
                                  filename=nameformat,
                                  message="when signal value is 0, se must be non-zero. please "
                                  + "use Jeffreys correction to generate an appropriate se"
                                  + " (see wikipedia.org/wiki/Binomial_proportion_confidence"
                                  + "_interval#Jeffreys_interval for details)"))
        elif not result_alt.empty:
            report.add_raised_error(
                ValidationFailure("check_se_0",
                                  filename=nameformat,
                                  message="se must be non-zero"))

        report.increment_total_checks()

    def check_bad_sample_size(self, df_to_test, nameformat, report):
        """
        Check sample sizes for validity.

        Arguments:
            - df_to_test: pandas dataframe of a single CSV of source data
            (one day-signal-geo_type combo)
            - nameformat: str CSV name; for example, "20200624_county_smoothed_nohh_cmnty_cli.csv"
            - report: ValidationReport; report where results are added

        Returns:
            - None
        """
        if self.params.missing_sample_size_allowed:
            result = df_to_test.query(
                '~(sample_size.isnull() | (sample_size >= @self.params.minimum_sample_size))')

            if not result.empty:
                report.add_raised_error(
                    ValidationFailure("check_n_missing_or_gt_min",
                                      filename=nameformat,
                                      message="sample size must be NA or >= "
                                              f"{self.params.minimum_sample_size}"))

            report.increment_total_checks()

        else:
            if df_to_test['sample_size'].isnull().values.any():
                report.add_raised_error(
                    ValidationFailure("check_n_missing",
                                      filename=nameformat,
                                      message="sample_size must not be NA"))

            report.increment_total_checks()

            # Find rows with sample size less than minimum allowed
            result = df_to_test.query(
                '(sample_size < @self.params.minimum_sample_size)')

            if not result.empty:
                report.add_raised_error(
                    ValidationFailure("check_n_gt_min",
                                      filename=nameformat,
                                      message="sample size must be >= "
                                              f"{self.params.minimum_sample_size}"))

            report.increment_total_checks()

    def check_duplicate_rows(self, data_df, filename, report):
        """
        Check if any rows are duplicated in a data set.

        Parameters
        ----------
        data_df: pd.DataFrame
            data to check
        filename: str
            name of file from which this data came
        report: ValidationReport
            report where results are added
        """
        is_duplicate = data_df.duplicated()
        if any(is_duplicate):
            report.add_raised_warning(
                ValidationFailure("check_duplicate_rows",
                                  filename=filename,
                                  message="Some rows are duplicated, which may indicate data "
                                          "integrity issues"))
        report.increment_total_checks()



# -*- coding: utf-8 -*-
"""Custom validator exceptions."""
import datetime as dt
from typing import Union, Optional

class APIDataFetchError(Exception):
    """Exception raised when reading API data goes wrong.

    Attributes:
        custom_msg -- parameters which caused the error
    """

    def __init__(self, custom_msg):
        """Initialize the APIDataFetchError with `custom_msg`."""
        self.custom_msg = custom_msg
        super().__init__(self.custom_msg)

    def __str__(self):
        """Extract the message."""
        return '{}'.format(self.custom_msg)

class ValidationFailure:
    """Structured report of single validation failure."""

    def __init__(self,
                 check_name: Optional[str]=None,
                 date: Optional[Union[str, dt.date]]=None,
                 geo_type: Optional[str]=None,
                 signal: Optional[str]=None,
                 message: str="",
                 filename: Optional[str]=None):
        """Initialize a ValidationFailure.

        Parameters
        ----------
        check_name: Optional[str]
            Name of check at which the failure happened.  A value of `None` is used to express all
            possible checks with a given `date`, `geo_type`, and/or `signal`.
        date: Optional[Union[str, dt.date]]
            Date corresponding to the data over which the failure happened.
            Strings are interpretted in ISO format ("YYYY-MM-DD").
            When `None`, the failure is not tied to any date.
        geo_type: Optional[str]
            Geo resolution of the data at which the failure occurred.
            When `None`, the failure is not tied to any geo resolution.
        signal: Optional[str]
            COVIDcast signal for which the failure occurred.
            When `None`, the failure is not tied to any signal.
        message: str
            Additional context about the failure
        filename: Optional[str]
            Filename from which the failing data came.
            When provided, any values passed to the `date`, `geo_type`, and `signal` parameters are
            ignored and instead derived from this parameter, under the assumption that the filename
            is formatted as "{date}_{geo_type}_{signal}.{extension}", where `date` is in the format
            "YYYYMMDD".

        Attributes
        ----------
        check_name: str
            See above.
        date: Optional[dt.date]
            See above.
        geo_type: Optional[str]
            See above.
        signal: Optional[str]
            See above.
        message: str
            See above.
        """
        self.check_name = check_name
        self.message = message
        if filename:
            pieces = filename.split(".")[0].split("_", maxsplit=2)
            assert len(pieces) == 3, '`filename` argument expected to be in "{date}_{geo_type}_'\
                                     '{signal}.{extension}" format'
            try:
                date = dt.datetime.strptime(pieces[0], "%Y%m%d").date()
            except ValueError as e:
                raise ValueError('date in `filename` must be in "YYYYMMDD" format') from e
            geo_type = pieces[1]
            signal = pieces[2]
        if isinstance(date, str):
            date = dt.date.fromisoformat(date)
        self.date = date
        self.geo_type = geo_type
        self.signal = signal

    def __eq__(self, other):
        """Compare this object with ValidationFailure `other` for equality.

        Two ValidationFailures are considered equal if their `check_name`, `date`, `geo_type`, and
        `signal` attributes correspondingly match.  A value of `None` in any of these attributes is
        considered to match any corresponding value on the other ValidationFailure.
        """
        def match_with_wildcard(x, y):
            """Determine if x and y are equal or None."""
            return x is None or y is None or x == y
        return match_with_wildcard(self.check_name, other.check_name) and\
            match_with_wildcard(self.date, other.date) and\
            match_with_wildcard(self.geo_type, other.geo_type) and\
            match_with_wildcard(self.signal, other.signal)

    def is_suppressed(self, suppressed_errors):
        """Determine whether the failure should be suppressed.

        Parameters
        ----------
        errors_to_suppress: List[ValidationFailure]
            set of data sources to ignore.
            Because we allow `None` values as wildcards, we cannot assume errors that are equal
            will hash to the same values.  Thus, we need to use a list rather than a set here.
        """
        return self in suppressed_errors

    def __str__(self):
        """Summarize context of failure."""
        date_str = "*" if self.date is None else self.date.isoformat()
        return f"{self.check_name} failed for {self.signal} at resolution {self.geo_type} on "\
               f"{date_str}: {self.message}"



"""Indicator running utilities."""
import argparse as ap
import importlib
import os
from typing import Any, Callable, Dict, Optional
import threading
from .archive import ArchiveDiffer, archiver_from_params
from .logger import get_structured_logger
from .utils import read_params, transfer_files, delete_move_files
from .validator.validate import Validator
from .validator.run import validator_from_params


Params = Dict[str, Any]

# Trivial function to use as default value for validator and archive functions.
NULL_FN = lambda x: None

def run_indicator_pipeline(indicator_fn:  Callable[[Params], None],
                            flash_fn: Callable[[Params], None] = NULL_FN,
                           validator_fn:  Callable[[Params], Optional[Validator]] = NULL_FN,
                           archiver_fn:  Callable[[Params], Optional[ArchiveDiffer]] = NULL_FN,
                             timer=1):
    """Run an indicator with its optional validation and archiving.

    Each argument to this function should itself be a function that will be passed a common set of
    parameters (see details below).  This parameter dictionary should have five subdictionaries
    keyed as "indicator", "validation", "archive", "flash", and "common" corresponding to parameters
    to be used in `indicator_fn`, `validator_fn`, `archiver_fn`, `flash_fn` and shared across
    functions, respectively.The timer function stops the flash process after a certain time.

    Arguments
    ---------
    indicator_fn: Callable[[Params], None]
        function that takes a dictionary of parameters and produces indicator output
    flash_fn: Callable[[Params], None]
        function that takes a dictionary of parameters and writes points of interest to the log.
    validator_fn: Callable[[Params], Optional[Validator]]
        function that takes a dictionary of parameters and produces the associated Validator or
        None if no validation should be performed.
    archiver_fn: Callable[[Params], Optional[ArchiveDiffer]]
        function that takes a dictionary of parameters and produces the associated ArchiveDiffer or
        None if no archiving should be performed.
    """
    params = read_params()
    logger = get_structured_logger(
        name=indicator_fn.__module__,
        filename=params["common"].get("log_filename", None),
        log_exceptions=params["common"].get("log_exceptions", True))

    #Get version and indicator name for startup
    ind_name = indicator_fn.__module__.replace(".run", "")
    #Check for version.cfg in indicator directory
    if os.path.exists("version.cfg"):
        with open("version.cfg") as ver_file:
            current_version = "not found"
            for line in ver_file:
                if "current_version" in line:
                    current_version = str.strip(line)
                    current_version = current_version.replace("current_version = ", "")
    #Logging - Starting Indicator
        logger.info(f"Started {ind_name} with covidcast-indicators version {current_version}")
    else: logger.info(f"Started {ind_name} without version.cfg")

    indicator_fn(params)
    validator = validator_fn(params)
    archiver = archiver_fn(params)

    if flash_fn:
        t = threading.Timer(timer, flash_fn(params))
        t.start()
        t.join(timer)
        if t.is_alive():
            t.cancel()
            t.join()


    if validator:
        validation_report = validator.validate()
        validation_report.log(logger)
        # Validators on dry-run always return success
        if not validation_report.success():
            delete_move_files()
    if (not validator or validation_report.success()):
        if archiver:
            archiver.run(logger)
        if "delivery" in params:
            transfer_files()


if __name__ == "__main__":
    parser = ap.ArgumentParser()
    parser.add_argument("indicator_name",
                        type=str,
                        help="Name of the Python package containing the indicator.  This package "
                             "must export a `run.run_module(params)` function.")
    args = parser.parse_args()
    indicator_module = importlib.import_module(args.indicator_name)
    flash_module = importlib.import_module('delphi_utils.flash_eval.run')
    run_indicator_pipeline(indicator_module.run.run_module,
                           flash_module.run_module,
                           validator_from_params,
                           archiver_from_params,
                           timer=600
                           )



# -*- coding: utf-8 -*-
"""Tools to validate CSV source data, including various check methods."""
from .datafetcher import load_all_files
from .dynamic import DynamicValidator
from .errors import ValidationFailure
from .report import ValidationReport
from .static import StaticValidator
from .utils import aggregate_frames, TimeWindow, end_date_helper

class Validator:
    """Class containing validation() function and supporting functions.

    Stores a list of all raised errors, and user settings.
    """

    def __init__(self, params):
        """
        Initialize object and set parameters.

        Arguments:
            - params: dictionary of user settings read from params.json file; if empty, defaults
                will be used
        """
        self.export_dir = params["common"]["export_dir"]

        assert "validation" in params, "params must have a top-level 'validation' object to run "\
            "validation"
        validation_params = params["validation"]
        suppressed_errors =  validation_params["common"].get('suppressed_errors', [])
        for entry in suppressed_errors:
            assert isinstance(entry, dict), "suppressed_errors must be a list of objects"
            assert set(entry.keys()).issubset(["check_name", "date", "geo_type", "signal"]),\
                'suppressed_errors may only have fields "check_name", "date", "geo_type", "signal"'

        self.suppressed_errors = [ValidationFailure(**entry) for entry in suppressed_errors]

        # Date/time settings
        validation_params["common"]["end_date"] = end_date_helper(validation_params)
        self.time_window = TimeWindow.from_params(validation_params["common"]["end_date"],
                                                  validation_params["common"]["span_length"])
        self.data_source = validation_params["common"].get("data_source", "")
        self.dry_run = validation_params["common"].get("dry_run", False)

        self.static_validation = StaticValidator(validation_params)
        self.dynamic_validation = DynamicValidator(validation_params)

    def validate(self):
        """
        Run all data checks.

        Arguments:
            - export_dir: path to data CSVs

        Returns:
            - ValidationReport collating the validation outcomes
        """
        report = ValidationReport(self.suppressed_errors, self.data_source, self.dry_run)
        frames_list = load_all_files(self.export_dir, self.time_window.start_date,
                                     self.time_window.end_date)
        self.static_validation.validate(frames_list, report)
        # Dynamic Validation only performed when frames_list is populated
        if len(frames_list) > 0:
            self.dynamic_validation.validate(aggregate_frames(frames_list), report)
        return report



# -*- coding: utf-8 -*-
"""Call the function run_module when executed.

This file indicates that running the module (`python -m delphi_validator`) will
call the function `run_module` found within the run.py file.
"""

from .run import run_module

run_module()



"""Functions for understanding and creating signal names."""


def add_prefix(signal_names, wip_signal, prefix="wip_"):
    """Add prefix to signal if there is a WIP signal.

    Parameters
    ----------
    signal_names: List[str]
        Names of signals to be exported
    prefix : "wip_"
        prefix for new/non public signals
    wip_signal : List[str] or bool
        a list of wip signals: [], OR
        all signals in the registry: True OR
        only signals that have never been published: False
    Returns
    -------
    List of signal names
        wip/non wip signals for further computation
    """
    if wip_signal is True:
        return [prefix + signal for signal in signal_names]
    if isinstance(wip_signal, list):
        make_wip = set(wip_signal)
        return [
            (prefix if signal in make_wip else "") + signal
            for signal in signal_names
        ]
    if wip_signal in {False, ""}:
        return signal_names
    raise ValueError("Supply True | False or '' or [] | list()")



# -*- coding: utf-8 -*-
"""Common Utility Functions to Support DELPHI Indicators."""

from __future__ import absolute_import

from .archive import ArchiveDiffer, GitArchiveDiffer, S3ArchiveDiffer
from .export import create_export_csv
from .utils import read_params

from .slack_notifier import SlackNotifier
from .logger import get_structured_logger
from .geomap import GeoMapper
from .smooth import Smoother
from .signal import add_prefix
from .nancodes import Nans
from .weekday import Weekday

__version__ = "0.3.9"



"""
Weekday effects (code from Aaron Rumack).

Created: 2020-05-06
"""
import cvxpy as cp
import numpy as np
from cvxpy.error import SolverError


class Weekday:
    """Class to handle weekday effects."""

    @staticmethod
    def get_params(data, denominator_col, numerator_cols, date_col, scales, logger):
        r"""Fit weekday correction for each col in numerator_cols.

        Return a matrix of parameters: the entire vector of betas, for each time
        series column in the data.
        """
        tmp = data.reset_index()
        denoms = tmp.groupby(date_col).sum()[denominator_col]
        nums = tmp.groupby(date_col).sum()[numerator_cols]

        # Construct design matrix to have weekday indicator columns and then day
        # indicators.
        X = np.zeros((nums.shape[0], 6 + nums.shape[0]))
        not_sunday = np.where(nums.index.dayofweek != 6)[0]
        X[not_sunday, np.array(nums.index.dayofweek)[not_sunday]] = 1
        X[np.where(nums.index.dayofweek == 6)[0], :6] = -1
        X[:, 6:] = np.eye(X.shape[0])

        npnums, npdenoms = np.array(nums), np.array(denoms)
        params = np.zeros((nums.shape[1], X.shape[1]))

        # Loop over the available numerator columns and smooth each separately.
        for i in range(nums.shape[1]):
            result = Weekday._fit(X, scales, npnums[:, i], npdenoms)
            if result is None:
                logger.error("Unable to calculate weekday correction")
            else:
                params[i,:] = result

        return params

    @staticmethod
    def _fit(X, scales, npnums, npdenoms):
        r"""Correct a signal estimated as numerator/denominator for weekday effects.

        The ordinary estimate would be numerator_t/denominator_t for each time point
        t. Instead, model

        log(numerator_t/denominator_t) = alpha_{wd(t)} + phi_t

        where alpha is a vector of fixed effects for each weekday. For
        identifiability, we constrain \sum alpha_j = 0, and to enforce this we set
        Sunday's fixed effect to be the negative sum of the other weekdays.

        We estimate this as a penalized Poisson GLM problem with log link. We
        rewrite the problem as

        log(numerator_t) = alpha_{wd(t)} + phi_t + log(denominator_t)

        and set a design matrix X with one row per time point. The first six columns
        of X are weekday indicators; the remaining columns are the identity matrix,
        so that each time point gets a unique phi. Using this X, we write

        log(numerator_t) = X beta + log(denominator_t)

        Hence the first six entries of beta correspond to alpha, and the remaining
        entries to phi.

        The penalty is on the L1 norm of third differences of phi (so the third
        differences of the corresponding columns of beta), to enforce smoothness.
        Third differences ensure smoothness without removing peaks or valleys.

        Objective function is negative mean Poisson log likelihood plus penalty:

        ll = (numerator * (X*b + log(denominator)) - sum(exp(X*b) + log(denominator)))
                / num_days
        """
        b = cp.Variable((X.shape[1]))

        lmbda = cp.Parameter(nonneg=True)
        lmbda.value = 10  # Hard-coded for now, seems robust to changes

        ll = (
            cp.matmul(npnums, cp.matmul(X, b) + np.log(npdenoms)) -
            cp.sum(cp.exp(cp.matmul(X, b) + np.log(npdenoms)))
        ) / X.shape[0]
        # L-1 Norm of third differences, rewards smoothness
        penalty = lmbda * cp.norm(cp.diff(b[6:], 3), 1) / (X.shape[0] - 2)
        for scale in scales:
            try:
                prob = cp.Problem(cp.Minimize((-ll + lmbda * penalty) / scale))
                _ = prob.solve()
                return b.value
            except SolverError:
                # If the magnitude of the objective function is too large, an error is
                # thrown; Rescale the objective function by going through loop
                continue
        return None

    @staticmethod
    def calc_adjustment(params, sub_data, cols, date_col):
        """Apply the weekday adjustment to a specific time series.

        Extracts the weekday fixed effects from the parameters and uses these to
        adjust the time series.

        Since

        log(numerator_t / denominator_t) = alpha_{wd(t)} + phi_t,

        we have that

        numerator_t / denominator_t = exp(alpha_{wd(t)}) exp(phi_t)

        and can divide by exp(alpha_{wd(t)}) to get a weekday-corrected ratio. In
        this case, we only divide the numerator, leaving the denominator unchanged
        -- this has the same effect.

        """
        tmp = sub_data.copy()

        for i, c in enumerate(cols):
            wd_correction = np.zeros((len(tmp[c])))

            for wd in range(7):
                mask = tmp[date_col].dt.dayofweek == wd
                wd_correction[mask] = tmp.loc[mask, c] / (
                    np.exp(params[i, wd]) if wd < 6 else np.exp(-np.sum(params[i, :6]))
                )
            tmp.loc[:, c] = wd_correction
        return tmp



# -*- coding: utf-8 -*-
"""Module to flagging interesting or unusual data points.

This file defines the functions that are made public by the module. As the
module is intended to be executed though the main method, these are primarily
for testing.
"""
from __future__ import absolute_import
from .constants import HTML_LINK



from setuptools import setup
from setuptools import find_packages

with open("README.md", "r") as f:
  long_description = f.read()

required = [
    "boto3",
    "covidcast",
    "cvxpy",
    "epiweeks",
    "freezegun",
    "gitpython",
    "mock",
    "moto",
    "numpy",
    "pandas>=1.1.0",
    "pydocstyle",
    "pylint==2.8.3",
    "pytest",
    "pytest-cov",
    "slackclient",
    "structlog",
    "xlrd"
]

setup(
    name="delphi_utils",
    version="0.3.9",
    description="Shared Utility Functions for Indicators",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="",
    author_email="",
    url="https://github.com/cmu-delphi/",
    install_requires=required,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(),
    package_data={'': ['data/20*/*.csv']}
)



# -*- coding: utf-8 -*-
"""Functions to call when running the tool.

This module should contain a function called `run_module`, that is executed
when the module is run with `python -m delphi_utils.validator`.
"""
import argparse as ap
from .. import read_params, get_structured_logger
from .validate import Validator


def run_module():
    """Run the validator as a module."""
    parser = ap.ArgumentParser()
    parser.add_argument("--dry_run", action="store_true", help="When provided, return zero exit"
                        " status irrespective of the number of failures")
    args = parser.parse_args()
    params = read_params()
    assert "validation" in params
    dry_run_param = params["validation"]["common"].get("dry_run", False)
    params["validation"]["common"]["dry_run"] = args.dry_run or dry_run_param
    validator = Validator(params)
    validator.validate().print_and_exit(
        get_structured_logger(__name__,
                              params["common"].get("log_filename", None)),
        not (args.dry_run or dry_run_param))


def validator_from_params(params):
    """Construct a validator from `params`.

    Arguments
    ---------
    params: Dict[str, Any]
        Dictionary of parameters
    """
    if "validation" in params:
        return Validator(params)
    return None
